#!/usr/bin/env python
# -*- coding: utf-8 -*-

###############################################################################
#
#  SOIF - Sofware for Optical Interferometry fitting
#  Copyright (C) 2016  Guillaume Schworer
#
#  This program is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
#  For any information, bug report, idea, donation, hug, beer, please contact
#    guillaume.schworer@obspm.fr
#
###############################################################################

"""
A package that includes oifits reading and fitting tools
"""

from time import gmtime as _gmtime


from .oidata import *
from .oifits import *
from .oigrab import *
from .oidataempty import *
from .oigrab import *
from .oipriors import *
from .oimodel import *
try:
    from .oifiting import *
except ImportError:
    pass
from .oiunitmodels import *
try:
    from .oiload import *
except ImportError:
    pass
try:
    from .oipriors import *
except:
    pass

from ._version import __version__, __major__, __minor__, __micro__

_disclaimer = """SOIF  Copyright (C) 2015-%s  Guillaume Schworer
This program comes with ABSOLUTELY NO WARRANTY.
This is free software, and you are welcome to redistribute it
under certain conditions.""" % _gmtime()[0]

print(_disclaimer)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

###############################################################################
#
#  SOIF - Sofware for Optical Interferometry fitting
#  Copyright (C) 2016  Guillaume Schworer
#
#  This program is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
#  For any information, bug report, idea, donation, hug, beer, please contact
#    guillaume.schworer@obspm.fr
#
###############################################################################


def doraise(obj, **kwargs):
    return bool(kwargs.pop('raiseError', getattr(obj, 'raiseError', True)))


def raiseIt(exc, raiseoupas, *args, **kwargs):
    exc = exc(*args, **kwargs)
    if raiseoupas:
        raise exc
    else:
        print("\033[31m{}\033[39m".format(exc.message))
        return True
    return False


class OIException(Exception):
    """
    Root for SOIF Exceptions, only used to trigger any soif
    errors, never raised
    """
    def _init(self, *args, **kwargs):
        self.args = [a for a in args] + [a for a in kwargs.values()]

    def __repr__(self):
        return repr(self.message)

    __str__ = __repr__


class NoTargetTable(OIException):
    """
    If the file has no OITARGET table
    """
    def __init__(self, src="", *args, **kwargs):
        self._init(src, *args, **kwargs)
        self.message = "There's no OI_TARGET table in the oifits file '{}'".format(src)


class NoWavelengthTable(OIException):
    """
    If the file has no OITARGET table
    """
    def __init__(self, src="", *args, **kwargs):
        self._init(src, *args, **kwargs)
        self.message = "There's no OI_WAVELENGTH table in the oifits file '{}'".format(src)


class ReadOnly(OIException):
    """
    If the parameter is read-only
    """
    def __init__(self, attr, *args, **kwargs):
        self._init(attr, *args, **kwargs)
        self.message = "Attribute '{}' is read-only".format(attr)


class InvalidDataType(OIException):
    """
    If the data type provided does not exist
    """
    def __init__(self, datatype, *args, **kwargs):
        self._init(datatype, *args, **kwargs)
        self.message = "Data type '{}' does not exist".format(datatype)


class HduDatatypeMismatch(OIException):
    """
    If the data type and the hdu provided do not match
    """
    def __init__(self, hduhead, datatype, *args, **kwargs):
        self._init(hduhead, datatype,
                                                  *args, **kwargs)
        self.message = "Data type '{}' and hdu with '{}' data do not match".format(datatype, hduhead)


class BadMaskShape(OIException):
    """
    If the mask shape does not match the data shape
    """
    def __init__(self, shape, *args, **kwargs):
        self._init(shape, *args, **kwargs)
        self.message = "Bad mask shape; it should be '{}'".format(shape)


class WrongData(OIException):
    """
    If the data provided has the wrong data type
    """
    def __init__(self, typ, *args, **kwargs):
        self._init(typ, *args, **kwargs)
        self.message = "Wrong data given, should be '{}'".format(typ)


class IncompatibleData(OIException):
    """
    If the data type and the hdu provided do not match
    """
    def __init__(self, typ1, typ2, *args, **kwargs):
        self._init(typ1, typ2, *args, **kwargs)
        self.message = "Can't merge '{}' and '{}'".format(typ1, typ2)


class NotADataHdu(OIException):
    """
    If the hdu provided does not contain data
    """
    def __init__(self, idx, src, *args, **kwargs):
        self._init(idx, src, *args, **kwargs)
        self.message = "Hdu index '{}' in file '{}' does not contain data".format(idx, src)


class NoSystematicsFit(OIException):
    """
    If the user did not set on the fit of systematics
    """
    def __init__(self, *args, **kwargs):
        self._init(*args, **kwargs)
        self.message = "You are not fitting systematics"


class NotCallable(OIException):
    """
    If the function is callable
    """
    def __init__(self, fct, *args, **kwargs):
        self._init(fct, *args, **kwargs)
        self.message = "'{}' should be callable".format(fct)


class NoEMCEE(OIException):
    """
    EMCEE is not installed
    """
    def __init__(self, *args, **kwargs):
        self._init(*args, **kwargs)
        self.message = "EMCEE is not installed"


class AllParamsSet(OIException):
    """
    When all parameters are set - nothing to run EMCEE on
    """
    def __init__(self, *args, **kwargs):
        self._init(*args, **kwargs)
        self.message = "All parameters are set, there is no reason for MCMC"


class InvalidBound(OIException):
    """
    Unitary model with invalid bounds
    """
    def __init__(self, name, param, vv, *args, **kwargs):
        self._init(name, param, vv, *args, **kwargs)
        self.message = "Object '{}', parameter '{}' has invalid bounds: {}".format(name, param, vv)


class NotFoundName(OIException):
    """
    Did not find the unitary model from its name
    """
    def __init__(self, name, *args, **kwargs):
        self._init(name, *args, **kwargs)
        self.message = "Object name '{}' not found".format(name)


class InalidUnitaryModel(OIException):
    """
    Did not find unitary model
    """
    def __init__(self, typ, *args, **kwargs):
        self._init(typ, *args, **kwargs)
        self.message = "Unitary model '{}' does not exist".format(typ)


class BusyName(OIException):
    """
    The unitary model name already exists
    """
    def __init__(self, name, *args, **kwargs):
        self._init(name, *args, **kwargs)
        self.message = "Object name '{}' already exists".format(name)


class BadParamsSize(OIException):
    """
    The input parameter 'params' has the wrong size
    """
    def __init__(self, size, *args, **kwargs):
        self._init(size, *args, **kwargs)
        self.message = "'params' size mismatch, should be '{:d}'".format(size)


class MasperpxMismatch(OIException):
    """
    The unitary model name already exists
    """
    def __init__(self, mpp1, mpp2, *args, **kwargs):
        self._init(mpp1, mpp2, *args, **kwargs)
        self.message = "Cannot rescale image, masperpx mismatch '{:.3f}' and '{:.3f}'".format(mpp1, mpp2)


class NoDataModel(OIException):
    """
    If there is no Oidata provided with the model and one tries to do
    stuff that is possible only with Oidata
    """
    def __init__(self, *args, **kwargs):
        self._init(*args, **kwargs)
        self.message = "There is no data in this model"


class ZeroErrorbars(OIException):
    """
    If there is some data errors are zero
    """
    def __init__(self, *args, **kwargs):
        self._init(*args, **kwargs)
        self.message = "Some data-errors are 0"

class shapeIssue(OIException):
    """
    If there is some data errors are zero
    """
    def __init__(self, uvshape, datashape, *args, **kwargs):
        self._init(uvshape, datashape, *args, **kwargs)
        self.message = "The data shape '{}' is not compatible with the (U,V) shape '{}'".format(uvshape, datashape)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

###############################################################################
#
#  SOIF - Sofware for Optical Interferometry fitting
#  Copyright (C) 2016  Guillaume Schworer
#
#  This program is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
#  For any information, bug report, idea, donation, hug, beer, please contact
#    guillaume.schworer@obspm.fr
#
###############################################################################


try:
    import astropy.io.fits as pf
except ImportError:
    import pyfits as pf

from .oifits import Oifits
from . import oiexception as exc
from . import core
np = core.np

__all__ = ['Oigrab']


class Oigrab(object):
    """
    Opens, reads and filters data contained in the OIFITS file ``src``.

    Args:
      * src (str): the path (relative or absolute) to the OIFITS file

    Kwargs:
      * raiseError (bool): if ``True``, raises errors, otherwise prints them

    Raises:
      * NoTargetTable: if the OIFITS file has no OI_TARGET table

    >>> import soif.oidata as oidata
    >>> data = oidata.Oigrab('./data/datafile.oifits')
    """
    def __init__(self, src, **kwargs):
        self._init(src=src, **kwargs)

    def _init(self, src, **kwargs):
        self.src = str(src)
        self.raiseError = bool(kwargs.pop('raiseError', True))
        hdus = pf.open(self.src)
        for item in hdus:
            if item.header.get('EXTNAME') == 'OI_TARGET':
                hdutgt = item
                break
        else:
            hdus.close()
            if exc.raiseIt(exc.NoTargetTable, self.raiseError, src=self.src):
                return
        self._targets = {}
        for ind, tgt in zip(hdutgt.data["TARGET_ID"], hdutgt.data["TARGET"]):
            self._targets[ind] = tgt
        if not core.hduWlindex(hdus):
            if exc.raiseIt(exc.NoWavelengthTable,
                           self.raiseError,
                           src=self.src):
                return
        # allwl = hdus[self._hduwlidx].data[core.KEYSWL['wl']]
        # self._wlmin = allwl.min()
        # self._wlmax = allwl.max()
        hdus.close()

    def _info(self):
        return "{}<SOIF File>{}\n File: '{}'".format(core.font.blue,
                                                     core.font.normal,
                                                     self.src)

    def __repr__(self):
        return self._info()

    __str__ = __repr__

    @property
    def targets(self):
        return self._targets

    @targets.setter
    def targets(self, value):
        exc.raiseIt(exc.ReadOnly, self.raiseError, attr="targets")

    def show_specs(self, ret=False, **kwargs):
        """
        Gets the target list and the data details from the OIFITS
        file.

        Args:
          * ret (bool): if ``True``, returns the information,
          otherwise prints it

        Returns:
          * a dictionary {'hdu index:info'} where info corresponds to
          a list of (Acquisition index, Target ID, MJD, N(UV), N(wl))
          tuples

        >>> import soif.oidata as oidata
        >>> data = oidata.Oigrab('./data/datafile.oifits')
        >>> data.showspecs()
        TARGETS:
        1: IRS_48
        2: Elia_2-15

        VIS2 [hdu=3]:
        Acq. Index | Target ID |      MJD       |  UVs | N wl
        -----------------------------------------------------
                 0 |         1 | 55636.3827746  |   21 |    1
                 1 |         2 | 55636.3827989  |   21 |    1
                 2 |         1 | 55636.3828232  |   21 |    1
        """
        hdus = pf.open(self.src)
        if not ret:
            print("TARGETS:")
            for ind, tgt in self.targets.items():
                print("{}: {}".format(ind, tgt))
        tgtlist = {}
        # formattitle = u"\n{} [hdu={:d}]: ({}{} µm)\n{}\n{}"
        formattitle = u"\n{} [hdu={:d}]:\n{}\n{}"
        formathead = u"{:^10} | {:^9} | {:^13} | {:^4} | {:^4}"
        formatline = u"{:>10} | {:>9} | {:>13} | {:>4} | {:>4}"
        for idx, item in enumerate(hdus):
            # if we have data in this header
            if core.hduToDataType(item) is not None:
                targetsortmjd, MJD, (ndata, nset, nunique, nholes, nwl) = \
                    core.gethduMJD(item, withdet=True)
                tgtlist[idx] = []
                if not ret:
                    print(formattitle.format(
                        core.hduToDataType(item),
                        idx,
                        # "{:.3f}".format(self._wlmin*1e6),
                        # "" if self._wlmax == self._wlmin \
                        #    else "-{:.3f}".format(self._wlmax*1e6),
                        formathead.format('Acq. Index', 'Target ID', 'MJD',
                                          'UVs', 'N wl'),
                        '-'*52))
                for tgtidx, sMJD in zip(targetsortmjd, MJD[targetsortmjd, 0]):
                    tgtfilter = slice(tgtidx*nunique, (tgtidx+1)*nunique)
                    tgtid = item.data['TARGET_ID'][tgtfilter]
                    info = (tgtidx, tgtid[0], sMJD, tgtid.size, nwl)
                    if not ret:
                        print(formatline.format(*info))
                    tgtlist[idx].append(info)
        hdus.close()
        if ret:
            return tgtlist

    def filtered(self, tgt=None, mjd=(None, None), hdus=(),
                      vis2=True, t3phi=True, t3amp=True, visphi=True,
                      visamp=True, verbose=False, **kwargs):
        """
        Give filtering parameters on the target name (OI_TARGET table),
        the observation wavelength (OI_WAVELENGTH table), the
        acquisition time [mjd_min, mjd_max], or the hdu index.
        Returns the data indices of the data matching all of these
        different filters. These lists are used to load the data within
        an Oidata object.
        Leave input parameter to 'None' to discard filtering on this
        particular parameter.
        """
        allhdus = pf.open(self.src)
        mjd = (float(mjd[0] if mjd[0] is not None else -np.inf),
               float(mjd[1] if mjd[1] is not None else np.inf))
        datayouwant = {'data': {'VIS2': bool(vis2),
                                'T3PHI': bool(t3phi),
                                'T3AMP': bool(t3amp),
                                'VISPHI': bool(visphi),
                                'VISAMP': bool(visamp)
                                }
                       }
        hdus = core.aslist(hdus)
        for idx, item in enumerate(allhdus):
            # do we want this hdu?
            if len(hdus) > 0 and idx not in hdus:
                    continue
            # is this hdu actual data?
            if core.hduToDataType(item) is not None:
                mjditem = core.gethduMJD(item)[1].ravel()
                filt = ((mjditem >= mjd[0]) & (mjditem <= mjd[1]))
                if tgt is not None:
                    filt = (filt & (item.data.field("TARGET_ID") == int(tgt)))
                if verbose:
                    print("hdu {:d}: {}:\n  {}/{}\n".format(
                                idx,
                                core.hduToDataType(item),
                                filt.sum(),
                                item.data["TARGET_ID"].size))
                if filt.any():
                    datayouwant[idx] = np.arange(item.data["TARGET_ID"].size)[filt]
        allhdus.close()
        return datayouwant

    def extract(self, tgt=None, mjd=(None, None), wl=(None, None), hdus=(),
                vis2=True, t3phi=True, t3amp=True, visphi=True, visamp=True,
                flatten=False, degree=True, significant_figures=5,
                erb_sigma=None, sigma_erb=None, systematic_prior=None,
                systematic_bounds=(), verbose=False, **kwargs):
        datayouwant = self.filtered(
                        tgt=tgt, mjd=mjd, vis2=vis2, hdus=hdus,
                        t3phi=t3phi, t3amp=t3amp, visphi=visphi, visamp=visamp,
                        verbose=verbose, **kwargs)
        return Oifits(src=self.src, datafilter=datayouwant, flatten=flatten,
                      degree=degree, significant_figures=significant_figures,
                      wl=wl, erb_sigma=erb_sigma, sigma_erb=sigma_erb,
                      systematic_prior=systematic_prior,
                      systematic_bounds=systematic_bounds, **kwargs)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

###############################################################################
#
#  SOIF - Sofware for Optical Interferometry fitting
#  Copyright (C) 2016  Guillaume Schworer
#
#  This program is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
#  For any information, bug report, idea, donation, hug, beer, please contact
#    guillaume.schworer@obspm.fr
#
###############################################################################


# import matplotlib.pyplot as _plt
try:
    import astropy.io.fits as pf
except ImportError:
    import pyfits as pf

from .oidataempty import OidataEmpty
from . import oiexception as exc
from . import core
np = core.np

__all__ = []


class Oidata(OidataEmpty):
    def __init__(self, src, hduidx, datatype, hduwlidx, indices=(),
                 wlindices=(), degree=True, flatten=False,
                 significant_figures=5, **kwargs):
        super(Oidata, self).__init__(datatype=datatype, **kwargs)
        self._input_src = [str(src)]
        self._input_hduidx = [int(hduidx)]
        self._input_hduwlidx = [int(hduwlidx)]
        hdus = pf.open(self._input_src[-1])
        hdu = hdus[self._input_hduidx[-1]]
        hduwl = hdus[self._input_hduwlidx[-1]]

        if core.DATAKEYSDATATYPE[self.datatype]['data'] \
           not in core.hduToColNames(hdu):
            if exc.raiseIt(exc.HduDatatypeMismatch,
                           self.raiseError,
                           hduhead=core.hduToDataType(hdu),
                           datatype=self.datatype):
                return

        self._input_degree = [bool(degree)]
        self._input_flatten = [bool(flatten)]
        self._input_significant_figures = [int(significant_figures)]
        self._input_indices = [list(indices)]
        self._input_wlindices = [list(wlindices)]

        self._has = True
        self._useit = True
        wlindices = slice(0, 10000000) if wlindices == () \
            else np.asarray(wlindices).ravel()
        indices = slice(0, 10000000) if indices == () \
            else np.asarray(indices).ravel()

        # attributes to be copy-pasted
        for key, vl in core.ATTRDATATYPE[self.datatype].items():
            setattr(self, '_'+key, vl)

        # wavelength attributes to be extracted out of OIFITS
        for key, vl in core.KEYSWL.items():
            setattr(self, '_'+key, hduwl.data[vl].ravel()[wlindices])
        self._wlsize = self._wl.size
        key = core.DATAKEYSDATATYPE[self.datatype]['data']
        self._datasize = hdu.data[key][indices].shape[0]

        # data attributes to be extracted out of OIFITS
        for key, vl in core.DATAKEYSDATATYPE[self.datatype].items():
            setattr(self,
                    '_'+key,
                    hdu.data[vl][indices]
                    .reshape((self._datasize, -1))[:, wlindices])
        # swaps Trues to Falses to have a good mask
        self.mask = np.logical_not(self._mask)

        # data attributes to be extracted out of OIFITS and replicated with wl
        for key, vl in core.UVKEYSDATATYPE[self.datatype].items():
            setattr(self,
                    '_'+key,
                    core.replicate(hdu.data[vl][indices].ravel(),
                                   (None, self._wlsize)))

        # done with the file
        hdus.close()

        # add the data dimension on wl-like attributes
        for key in core.KEYSWL.keys():
            setattr(self,
                    "_"+key,
                    core.replicate(self[key],
                                   (self._datasize, None)))

        # combine the UV coordinates of T3
        if self.is_t3:
            self._u = np.concatenate((np.expand_dims(self._u1, -1),
                                      np.expand_dims(self._u2, -1),
                                      np.expand_dims(-self._u1-self._u2, -1)),
                                     axis=-1)
            self._v = np.concatenate((np.expand_dims(self._v1, -1),
                                      np.expand_dims(self._v2, -1),
                                      np.expand_dims(-self._v1-self._v2, -1)),
                                     axis=-1)
            # delete temporary attributes
            for key in core.UVKEYSDATATYPE[self.datatype].keys():
                delattr(self, "_"+key)
            # add the data dimension on wl-like attributes
            for key in core.KEYSWL.keys():
                setattr(self,
                        "_"+key,
                        core.replicate(self[key], (None, 3)))

        # check shape sanity - might be raised in case there are several wl
        # tables in the oifits. This feature is not covered by this library
        extra_dim = (3,) if self.is_t3 else ()
        if not self._u.shape == self._data.shape + extra_dim:
            if exc.raiseIt(exc.shapeIssue,
                           self.raiseError,
                           uvshape=self._u.shape[:2],
                           datashape=self._data.shape):
                return

        # convert to radian if needed
        if self.is_angle and degree:
            self._data = self._data*core.DEG2RAD
            self._error = self._error*core.DEG2RAD

        self.significant_figures = min(8, max(1, int(significant_figures)))
        self._flat = bool(flatten)
        if self._flat:
            self.flatten()
        else:
            self.update()

    def _info(self):
        return str(u"{} data, shape: {}, wl: {:.2f}{} \xb5m".format(
            self.datatype,
            core.maskedshape(self.shapedata,
                             np.logical_not(self.mask).sum()),
            self._wlmin*1e6,
            u" to {:.2f}".format(self._wlmax*1e6) if self._wlspan != 0 else u""
            ).encode('utf-8'))

    @property
    def data(self):
        if self._use_mask:
            return self._data[self.mask]
        else:
            return self._data

    @data.setter
    def data(self, value):
        exc.raiseIt(exc.ReadOnly, self.raiseError, attr="data")

    @property
    def error(self):
        if self._use_mask:
            return self._error[self.mask]
        else:
            return self._error

    @error.setter
    def error(self, value):
        exc.raiseIt(exc.ReadOnly, self.raiseError, attr="error")

    @property
    def mask(self):
        return self._mask

    @mask.setter
    def mask(self, value):
        if value is True or value is None:  # removes the mask
            self._use_mask = False
            self._mask = np.ones(self._data.shape, dtype=bool)
        elif value is False:  # set up full masking
            self._use_mask = True
            self._mask = np.zeros(self._data.shape, dtype=bool)
        else:  # convert value to proper bool mask
            value = np.asarray(value, dtype=bool)
            if value.shape != self._data.shape:
                if exc.raiseIt(exc.BadMaskShape,
                               self.raiseError,
                               shape=self._data.shape):
                    return False
            self._use_mask = not value.all()
            self._mask = value

    @property
    def u(self):
        if self._use_mask:
            return self._u[self.mask]
        else:
            return self._u

    @u.setter
    def u(self, value):
        exc.raiseIt(exc.ReadOnly, self.raiseError, attr="u")

    @property
    def v(self):
        if self._use_mask:
            return self._v[self.mask]
        else:
            return self._v

    @v.setter
    def v(self, value):
        exc.raiseIt(exc.ReadOnly, self.raiseError, attr="v")

    @property
    def wl(self):
        if self._use_mask:
            return self._wl[self.mask]
        else:
            return self._wl

    @wl.setter
    def wl(self, value):
        exc.raiseIt(exc.ReadOnly, self.raiseError, attr="wl")

    @property
    def wl_d(self):
        if self._use_mask:
            return self._wl_d[self.mask]
        else:
            return self._wl_d

    @wl_d.setter
    def wl_d(self, value):
        exc.raiseIt(exc.ReadOnly, self.raiseError, attr="wl_d")

    @property
    def bl(self):
        if self._use_mask:
            return self._bl[self.mask]
        else:
            return self._bl

    @bl.setter
    def bl(self, value):
        exc.raiseIt(exc.ReadOnly, self.raiseError, attr="bl")

    @property
    def pa(self):
        if self._use_mask:
            return self._pa[self.mask]
        else:
            return self._pa

    @pa.setter
    def pa(self, value):
        exc.raiseIt(exc.ReadOnly, self.raiseError, attr="pa")

    @property
    def blwl(self):
        if self._use_mask:
            return self._blwl[self.mask]
        else:
            return self._blwl

    @blwl.setter
    def blwl(self, value):
        exc.raiseIt(exc.ReadOnly, self.raiseError, attr="blwl")

    @property
    def shapedata(self):
        return self.data.shape

    @shapedata.setter
    def shapedata(self, value):
        exc.raiseIt(exc.ReadOnly, self.raiseError, attr="shapedata")

    @property
    def shapeuv(self):
        return self.u.shape

    @shapeuv.setter
    def shapeuv(self, value):
        exc.raiseIt(exc.ReadOnly, self.raiseError, attr="shapeuv")

    @property
    def is_angle(self):
        return self._is_angle

    @is_angle.setter
    def is_angle(self, value):
        exc.raiseIt(exc.ReadOnly, self.raiseError, attr="is_angle")

    @property
    def is_t3(self):
        return self._is_t3

    @is_t3.setter
    def is_t3(self, value):
        exc.raiseIt(exc.ReadOnly, self.raiseError, attr="is_t3")

    @property
    def flat(self):
        return self._flat

    @flat.setter
    def flat(self, value):
        exc.raiseIt(exc.ReadOnly, self.raiseError, attr="flat")

    def flatten(self, **kwargs):
        if self.is_t3:
            for key in core.KEYSUV:
                setattr(self, "_"+key, self[key].reshape((-1, 3)))
            for key in core.KEYSDATA:
                setattr(self, "_"+key, self[key].ravel())
        else:
            for key in core.KEYSDATA + core.KEYSUV:
                setattr(self, "_"+key, self[key].ravel())
        self.update()
        self._flat = True

    def _addData(self, data, flatten=True, **kwargs):
        if isinstance(data, OidataEmpty):
            return  # trivial, nothing to add
        if not isinstance(data, Oidata):
            if exc.raiseIt(exc.WrongData,
                           exc.doraise(self, **kwargs),
                           typ='Oidata'):
                return False
        if self.datatype != data.datatype:
            if exc.raiseIt(exc.IncompatibleData,
                           exc.doraise(self, **kwargs),
                           typ1=self.datatype,
                           typ2=data.datatype):
                return False
        # do we flatten it?
        if flatten or self.shapedata[-1] != data.shapedata[-1]:
            self.flatten()
            data.flatten()
        # concatenate data
        for key in core.KEYSDATA + core.KEYSUV:
            setattr(self,
                    "_"+key,
                    np.concatenate((self[key], data[key]), axis=0))
        # update input keys
        for key in core.INPUTSAVEKEY:
            setattr(self,
                    "_input_"+key,
                    self["input_"+key] + data["input_"+key])
        # update the data
        self.update()

    def update(self, **kwargs):
        """
        Given u, v, wl and flag information as object properties, this
        function updates the Oidata object: the data masking (from the
        new mask property) and the bl, pa, blwl properties (from u, v
        and wl properties)
        """
        self._u = core.round_fig(self._u,
                                 self.significant_figures)
        self._v = core.round_fig(self._v,
                                 self.significant_figures)
        self._wl = core.round_fig(self._wl,
                                  self.significant_figures)
        self._wl_d = core.round_fig(self._wl_d,
                                    self.significant_figures)
        if (self.error == 0).any():
            if exc.raiseIt(exc.ZeroErrorbars, exc.doraise(self, **kwargs)):
                return False
        self._invvar = 1./self.error**2
        self._bl = core.round_fig(np.hypot(self['v'], self['u']),
                                  self.significant_figures)
        self._pa = core.round_fig(np.arctan2(self['v'], self['u']),
                                  self.significant_figures)
        self._blwl = core.round_fig(self['bl']/self['wl'],
                                    self.significant_figures)
        self._wlmin = self.wl.min()
        self._wlmax = self.wl.max()
        self._wlspan = self._wlmax - self._wlmin



#!/usr/bin/env python
# -*- coding: utf-8 -*-

###############################################################################
#
#  SOIF - Sofware for Optical Interferometry fitting
#  Copyright (C) 2016  Guillaume Schworer
#
#  This program is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
#  For any information, bug report, idea, donation, hug, beer, please contact
#    guillaume.schworer@obspm.fr
#
###############################################################################

from time import strftime
from time import time

try:
    import astropy.io.fits as pf
except ImportError:
    import pyfits as pf

from . import oipriors
from . import oiexception as exc
from . import core
np = core.np


__all__ = ['Oimainobject']


class Oimainobject(object):
    _keys = []

    def __init__(self, name, priors={}, bounds={}, verbose=False, *args, **kwargs):
        super(Oimainobject, self).__init__(*args, **kwargs)
        self.raiseError = bool(kwargs.pop('raiseError', True))
        self.name = str(name)
        self._nkeys = len(self._keys)
        self._pkeys = []
        self._nparams = 0
        self._vkeys = []
        self._pmask = np.zeros(self._nkeys, dtype=bool)
        self._vmask = np.zeros(self._nkeys, dtype=bool)
        for item in bounds.keys():
            if item not in self._keys:
                print("{}WARNING{}: Ignoring unknown bounds '{}' for object '{}'.".format(
                                core.font.orange,
                                core.font.normal,
                                item,
                                name))
        for item in priors.keys():
            if item not in self._keys:
                print("{}WARNING{}: Ignoring unknown prior '{}' for object '{}'.".format(
                                core.font.orange,
                                core.font.normal,
                                item,
                                name))
        for i, item in enumerate(self._keys):
            found = False
            if item in priors.keys():
                found = True
                setattr(self, item, float(priors[item]))
                setattr(self, "_"+item, float(priors[item]))
            if item in bounds.keys():
                vals = bounds[item]
                found = True
                if len(vals) < 2:  # error if less than 2 bounds
                    if exc.raiseIt(exc.InvalidBound, self.raiseError, name=name, param=item, vv=str(vals)):
                        return
                # set basic parameters
                setattr(self, item+"_bounds", list(map(float, vals[:2])))
                prior_range = np.abs(getattr(self, item+"_bounds")[1]-getattr(self, item+"_bounds")[0])*1.
                allkwargs = {'prior_bounds': getattr(self, item+"_bounds"), 'prior_range': prior_range, 'prior_invrange': 1./prior_range, 'prior_lninvrange': -np.log(prior_range)}
                if len(vals)>=4:
                    if not isinstance(vals[3], dict):
                        raise Exception("For object '{}', the fourth argument '{}' in prior definition must be a dictionary.".format(item, vals[3]))
                    allkwargs.update(vals[3])
                setattr(self, item+"_prior_kwargs", allkwargs)
                if len(vals)==2:
                    if verbose:
                        print("{}WARNING{}: No prior probability function found for parameter '{}', object '{}'. Assuming Uniform.".format(core.font.orange, core.font.normal, item, name))
                    setattr(self, item+"_prior_lnfunc", core.PriorWrapper(oipriors.lnuniform, makenorm=True, makeunlog=False, fct_log=True, **allkwargs)) # default is uniform on the range
                    setattr(self, item+"_prior_func", core.PriorWrapper(oipriors.uniform, makenorm=True, makeunlog=False, fct_log=False, **allkwargs)) # default is uniform on the range
                if len(vals)>=3:
                    if not callable(vals[2]):
                        raise Exception("For object '{}', the third argument '{}' in bounds definition must be callable.".format(item, vals[2]))
                    setattr(self, item+"_prior_lnfunc", core.PriorWrapper(vals[2], makenorm=True, makeunlog=False, fct_log=True, **allkwargs))
                if len(vals) in [3, 4]:
                    if verbose:
                        print("{}WARNING{}: No bounds probability function found for parameter '{}', object '{}'. Initial values for this parameter will be obtained from the exponentiation of the log-probability function.".format(core.font.orange, core.font.normal, item, name))
                    setattr(self, item+"_prior_func", core.PriorWrapper(vals[2], makenorm=True, makeunlog=True, fct_log=False, **allkwargs))
                if len(vals)==5:
                    if not callable(vals[4]):
                        raise Exception("For object '{}', the fifth argument '{}' in prior definition must be callable.".format(item, vals[4]))
                    setattr(self, item+"_prior_func", core.PriorWrapper(vals[4], makenorm=True, makeunlog=False, fct_log=False, **allkwargs))
                # pre-computing of P0 stuff
                x = getattr(getattr(self, item+"_prior_func"), 'x')
                setattr(self, "_"+item+"_P0", core.random_custom_pdf(x, getattr(self, item+"_prior_func")(x), size=False, renorm=True))
                self._pkeys.append(item)
                self._pmask[i] = True
                self._nparams += 1
            else:
                self._vkeys.append(item)
                self._vmask[i] = True
            if not found:
                exc.raiseIt(exc.InvalidBound, self.raiseError, name=name, param=item, vv="None")
                # force 0.0 to prior
                setattr(self, item, 0.0)
                setattr(self, "_"+item, 0.0)
                return

    @property
    def typ(self):
        return self.__class__.__name__
    @typ.setter
    def typ(self, value):
        exc.raiseIt(exc.ReadOnly, self.raiseError, attr="typ")

    def _info(self):
        anyparam = len(self._pkeys)>0
        anyvalue = len(self._vkeys)>0
        ret = "{}<{}>{} '{}': {:d} params".format(core.font.yellow, self.typ, core.font.normal, self.name,
            self._nparams)
        if anyparam:
            ret += " (" + ", ".join([item for item in self._pkeys]) + ")"
        if anyvalue:
            ret += "; " + ", ".join([item+"="+str(getattr(self, item)) for item in self._vkeys])
        return ret
    def __repr__(self):
        return self._info()
    def __str__(self):
        return self._info()

    def show(self):
        ret = "{}<{}>{} '{}': {:d} params".format(core.font.yellow, self.typ, core.font.normal, self.name, self._nparams)
        if len(self._vkeys)>0:
            ret += "\nValues:"
            for item in self._vkeys:
                ret += "\n  {}{}{}={}".format(core.font.orange, item, core.font.normal, getattr(self, item))
        if len(self._pkeys)>0:
            ret += "\nParams:"
            for item in self._pkeys:
                ret += "\n  {}{}{}={} in [{}, {}]".format(core.font.orange, item, core.font.normal, getattr(self, item, "."), getattr(self, item+"_bounds")[0], getattr(self, item+"_bounds")[1])
        print(ret)

    @classmethod
    def keys(cls):
        return getattr(cls, '_keys', [])

    def to_radec(self):
        """
        Returns (ra, dec) in radian from sep and pa
        """
        seprad = self.sep*core.MAS2RAD
        return seprad*np.sin(self.pa), seprad*np.cos(self.pa)

    def to_pospx(self, sepmax, nbpts, integer=False):
        norm = nbpts/(2.*sepmax)
        ra, dec = self.sep*np.asarray([np.sin(self.pa), np.cos(self.pa)])*norm+0.5*(nbpts-1)
        if integer: ra, dec = np.round([ra, dec]).astype(int)
        return ra, dec

    def _shearCoord(self, x, y, fourier=False):
        if fourier:
            rat = 1./self.rat
        else:
            rat = self.rat
        cth = np.cos(self.th)
        sth = np.sin(self.th)
        return (x*cth-y*sth)*rat, x*sth+y*cth

    def oscil(self, u, v, wl):
        if self.sep!=0:
            ret = np.zeros(u.shape, dtype=complex)
            dra, ddec = self.to_radec()
            # analytical fourier transform for a point source
            oscil = -core.DEUXPI*(u*dra + v*ddec) / wl
            ret.real = np.cos(oscil)
            ret.imag = np.sin(oscil)
            return ret
        else:
            return np.ones(u.shape, dtype=complex)

    @property
    def params(self):
        """
        Returns a list of the parameters values, according to the parameter keys order
        """
        return [getattr(self, item) if hasattr(self, item) else np.mean(getattr(self, item+"_bounds")) for item in self._pkeys]
    @params.setter
    def params(self, value):
        exc.raiseIt(exc.ReadOnly, self.raiseError, attr="params")

    def setParams(self, params, priors=False):
        """
        Affects to the object the registered values from params, according the to parameter keys order
        """
        for ind, arg in enumerate(self._pkeys):
            setattr(self, arg, params[ind])
        if priors:
            for ind, arg in enumerate(self._pkeys):
                setattr(self, "_"+arg, params[ind])

    def getP0(self):
        """
        Returns a list of initial values for each parameter in the object, according the to parameter keys order
        """
        ret = []
        randomizer = core.gen_generator()
        for item in self._pkeys:
            ret.append(np.clip(getattr(self, "_"+item+"_P0")(randomizer.uniform()), a_min=getattr(self, item+"_bounds")[0], a_max=getattr(self, item+"_bounds")[1]))
        return ret

    def compVis(self, oidata, params=None, flat=False):
        """
        Calculates the complex visibilities of the object
        """
        if params is not None: self.setParams(params)
        if flat:
            return self._calcCompVis(u=oidata.uvwl['u'], v=oidata.uvwl['v'], wl=oidata.uvwl['wl'], blwl=oidata.uvwl['blwl'], oidata=oidata)
        else:
            return oidata.remorph(self._calcCompVis(u=oidata.uvwl['u'], v=oidata.uvwl['v'], wl=oidata.uvwl['wl'], blwl=oidata.uvwl['blwl'], oidata=oidata))

    def save(self, filename, append=False, clobber=False):
        ext = '.oif.fits'
        if filename.find(ext)==-1: filename += ext
        if append:
            hdulist = pf.open(filename, mode='append')
        else:
            hdulist = pf.HDUList()
        hdu = pf.PrimaryHDU()
        hdu.header.set('DATE', strftime('%Y%m%dT%H%M%S'), comment='Creation Date')
        hdu.header.set('EXT', 'OBJ', comment='Type of information in the HDU')
        hdu.header.set('NAME', str(self.name), comment='Name of the object')
        hdu.header.set('TYP', str(self.typ), comment='Unit-model of the object')
        hdu.header.set('NVALUE', len(self._vkeys), comment='Number of set parameters in the model')
        hdu.header.set('NPARAM', len(self._pkeys), comment='Number of free parameters in the model')
        hdu.header.set('NSAVE', len(getattr(self, '_save', {})), comment='Number of additional parameters in the model')
        for i, k in enumerate(self._vkeys):
            hdu.header.set('VALUE'+str(i), k, comment='Name of the set parameter '+str(i))
            hdu.header.set(k,  getattr(self, "_"+k), comment='Value of the set parameter '+str(i))
        for i, k in enumerate(self._pkeys):
            hdu.header.set('PARAM'+str(i), k, comment='Name of the free parameter '+str(i))
            hdu.header.set(k,  getattr(self, "_"+k, 'NONE'), comment='Prior of the free parameter '+str(i))
            hdu.header.set(k+'L',  getattr(self, k+"_bounds")[0], comment='Value of the lower prior-bound ')
            hdu.header.set(k+'H',  getattr(self, k+"_bounds")[1], comment='Value of the higher prior-bound ')
        for i, k in enumerate(getattr(self, '_save', [])):
            hdu.header.set('SAVE'+str(i), k, comment='Name of the additional parameter '+str(i))
            if not isinstance(getattr(self, k), (np.ndarray, list, tuple)):
                hdu.header.set(k,  getattr(self, k), comment='Value of the additional parameter '+str(i))
            else:
                hdu.header.set(k, 'ARRAY', comment='Value of the additional parameter '+str(i))
                key = 'K'+str(int(time()/100%1*10**7))
                hdu.header.set('KEY'+str(i), key, comment='Unique key to find the parameter value '+str(i))
                paramhdu = pf.PrimaryHDU()
                paramhdu.header.set('DATE', strftime('%Y%m%dT%H%M%S'), comment='Creation Date')
                paramhdu.header.set('NAME', str(self.name), comment='Name of the object')
                paramhdu.header.set('TYP', str(self.typ), comment='Unit-model of the object')
                paramhdu.header.set('SAVE'+str(i), k, comment='Name of the additional parameter '+str(i))
                paramhdu.header.set('EXT', key, comment='Type of information in the HDU : unique key')
                paramhdu.data = np.asarray(getattr(self, k))
                hdulist.append(paramhdu)

        hdu.header.add_comment('Written by Guillaume SCHWORER')
        hdulist.append(hdu)
        
        if append:
            hdulist.flush()
            hdulist.close()
        else:
            hdulist.writeto(filename, clobber=clobber)
        return filename



#!/usr/bin/env python
# -*- coding: utf-8 -*-

###############################################################################
#  
#  SOIF - Sofware for Optical Interferometry fitting
#  Copyright (C) 2016  Guillaume Schworer
#  
#  This program is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#  
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#  
#  You should have received a copy of the GNU General Public License
#  along with this program.  If not, see <http://www.gnu.org/licenses/>.
#  
#  For any information, bug report, idea, donation, hug, beer, please contact
#    guillaume.schworer@obspm.fr
#
###############################################################################

import numpy as np

__all__ = []


def lnuniform(x, prior_lninvrange=0, *args, **kwargs):
    """
    Returns the log-probability of drawing any value in the range of the prior
    """
    if np.size(x)==1: return prior_lninvrange
    return np.ones(np.shape(x))*prior_lninvrange

def uniform(x, prior_invrange=0, *args, **kwargs):
    """
    Returns the probability of drawing any value in the range of the prior
    """
    if np.size(x)==1: return prior_invrange
    return np.ones(np.shape(x))*prior_invrange


def lnnormal(x, x0=0., sigma=1., *args, **kwargs):
    """
    Returns the probability of drawing x
    """
    return -0.5*((x-x0)/sigma)**2

def normal(x, x0=0., sigma=1., *args, **kwargs):
    """
    Returns the probability of drawing x
    """
    return np.exp(-0.5*((x-x0)/sigma)**2)


def lnnormalbumpy(x, x0=0., rangeshrink=10, *args, **kwargs):
    """
    Returns the probability of drawing x
    """
    return np.exp(-0.5*((x-x0)*pctrange*kwargs['prior_invrange'])**2)

def normalbumpy(x, x0=0., rangeshrink=10, *args, **kwargs):
    """
    Returns the probability of drawing x
    """
    return np.exp(np.exp(-0.5*((x-x0)*pctrange*kwargs['prior_invrange'])**2))


def lntriangle(x, x0=0, slope=4, *args, **kwargs):
    """
    Returns the log-probability of drawing any value in the range of the prior
    """
    return -slope*np.abs(x-x0)*kwargs['prior_invrange']

def triangle(x, x0=0, slope=4, *args, **kwargs):
    """
    Returns the probability of drawing any value in the range of the prior
    """
    return np.exp(-slope*np.abs(x-x0)*kwargs['prior_invrange'])



#!/usr/bin/env python
# -*- coding: utf-8 -*-

###############################################################################
#
#  SOIF - Sofware for Optical Interferometry fitting
#  Copyright (C) 2016  Guillaume Schworer
#
#  This program is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
#  For any information, bug report, idea, donation, hug, beer, please contact
#    guillaume.schworer@obspm.fr
#
###############################################################################


from scipy.special import j1 as airyJ1
from scipy.signal import fftconvolve  # it is used, just not in that file
import numpy as np
from time import time as timetime
from multiprocessing import current_process as multiprocessingcurrent_process
from scipy.interpolate import interp1d as scipyinterpolateinterp1d
from scipy.integrate import cumtrapz as scipyintegratecumtrapz
from matplotlib.cm import get_cmap as cmget_cmap
from matplotlib.cm import ScalarMappable as cmScalarMappable
from matplotlib.pyplot import Normalize as matplotlibpyplotNormalize
from matplotlib.pyplot import subplots as matplotlibpyplotsubplots
from sys import maxsize as sysmaxint
sysmaxint = min(2**32, sysmaxint)

__all__ = []


DATATYPEEXTNAMES = {'OI_T3': 'T3',
                    'OI_VIS2': 'VIS2',
                    'OI_VIS': 'VIS'
                    }
DATAEXTNAMES = {'OI_T3': 'T3PHI',
                'OI_VIS2': 'VIS2DATA',
                'OI_VIS': 'VISPHI'
                }
ALLDATAEXTNAMES = {'OI_T3': ['T3AMP', 'T3PHI'],
                   'OI_VIS2': ['VIS2'],
                   'OI_VIS': ['VISPHI', 'VISAMP']
                   }
DATAKEYSDATATYPE = {'T3AMP': {'data': 'T3AMP',
                              'error': 'T3AMPERR',
                              'mask': 'FLAG'
                              },
                    'T3PHI': {'data': 'T3PHI',
                              'error': 'T3PHIERR',
                              'mask': 'FLAG'
                              },
                    'VIS2': {'data': 'VIS2DATA',
                             'error': 'VIS2ERR',
                             'mask': 'FLAG'
                             },
                    'VISPHI': {'data': 'VISPHI',
                               'error': 'VISPHIERR',
                               'mask': 'FLAG'},
                    'VISAMP': {'data': 'VISAMP',
                               'error': 'VISAMPERR',
                               'mask': 'FLAG'
                               }
                    }
UVKEYSDATATYPE = {'T3AMP': {'u1': 'U1COORD',
                            'v1': 'V1COORD',
                            'u2': 'U2COORD',
                            'v2': 'V2COORD'
                            },
                  'T3PHI': {'u1': 'U1COORD',
                            'v1': 'V1COORD',
                            'u2': 'U2COORD',
                            'v2': 'V2COORD'
                            },
                  'VIS2': {'u': 'UCOORD',
                           'v': 'VCOORD'
                           },
                  'VISPHI': {'u': 'UCOORD',
                             'v': 'VCOORD'
                             },
                  'VISAMP': {'u': 'UCOORD',
                             'v': 'VCOORD'
                             }
                  }
ATTRDATATYPE = {'T3AMP': {'is_t3': True,
                          'is_angle': False
                          },
                'T3PHI': {'is_t3': True,
                          'is_angle': True
                          },
                'VIS2': {'is_t3': False,
                         'is_angle': False
                         },
                'VISPHI': {'is_t3': False,
                           'is_angle': True
                           },
                'VISAMP': {'is_t3': False,
                           'is_angle': False
                           }
                }
DATAKEYSLOWER = [item.lower() for item in ATTRDATATYPE.keys()]
DATAKEYSUPPER = [item.upper() for item in ATTRDATATYPE.keys()]


def abs2(ar):
    return np.abs(ar)**2

FCTVISCOMP = {'VIS2': abs2,
              'T3PHI': np.angle,
              'T3AMP': np.abs,
              'VISPHI': np.angle,
              'VISAMP': np.abs
              }

KEYSWL = {'wl': 'EFF_WAVE', 'wl_d': 'EFF_BAND'}
KEYSDATA = ['data', 'error', 'mask']
KEYSUV = ['v', 'u', 'wl', 'wl_d']
INPUTSAVEKEY = ['src',
                'hduidx',
                'hduwlidx',
                'degree',
                'flatten',
                'significant_figures',
                'indices',
                'wlindices'
                ]

NHOLES = np.arange(2, 100)
TRIHOLESTAB = NHOLES*(NHOLES-1)*(NHOLES-2)/6
BLHOLESTAB = NHOLES*(NHOLES-1)/2

OLDNUMPY = int(np.version.version.split('.')[1]) < 10
# multiply mili-arc-second (mas) with that and you get radian
MAS2RAD = 4.84813681109536e-9
RAD2MAS = 1/MAS2RAD
# multiply arc second with that and you get radian
ASEC2RAD = 4.84813681109536e-6
RAD2ASEC = 1/ASEC2RAD
# multiply degrees with that and you get radian
DEG2RAD = 0.017453292519943295
RAD2DEG = 1/DEG2RAD
DEUXPI = 2*np.pi
PISQRT2 = np.pi*np.sqrt(2)
LNDEUXPI = np.log(2*np.pi)
PISURDEUX = 0.5*np.pi
SECPERDAY = 86400.


ALL_FILTERS = {'U': {'mean_wl':         3.60e-7,
                     'delta_wl':        5.40e-8,
                     'flux_jansky':     1810.,
                     'flux_wm2m':       4.186e-2
                     },
               'B': {'mean_wl':         4.40e-7,
                     'delta_wl':        9.68e-8,
                     'flux_jansky':     4260.,
                     'flux_wm2m':       6.596e-2
                     },
               'V': {'mean_wl':         5.50e-7,
                     'delta_wl':        8.80e-8,
                     'flux_jansky':     3540.,
                     'flux_wm2m':       3.508e-2
                     },
               'R': {'mean_wl':         7.00e-7,
                     'delta_wl':        1.47e-7,
                     'flux_jansky':     2880.,
                     'flux_wm2m':       1.762e-2
                     },
               'I': {'mean_wl':         9.00e-7,
                     'delta_wl':        1.50e-7,
                     'flux_jansky':     2250.,
                     'flux_wm2m':       8.327e-3
                     },
               'J': {'mean_wl':         1.25e-6,
                     'delta_wl':        2.02e-7,
                     'flux_jansky':     1670.,
                     'flux_wm2m':       3.204e-3
                     },
               'H': {'mean_wl':         1.65e-6,
                     'delta_wl':        3.68e-7,
                     'flux_jansky':     981.,
                     'flux_wm2m':       1.080e-3
                     },
               'K': {'mean_wl':         2.20e-6,
                     'delta_wl':        5.11e-7,
                     'flux_jansky':     620.,
                     'flux_wm2m':       3.840e-4
                     }
               }


def _gen_lam_for_filter(filt, nbpts=10):
    return np.logspace(np.log10(item['start_wl']),
                       np.log10(item['end_wl']),
                       10)


for idx, item in ALL_FILTERS.items():
    item['start_wl'] = item['mean_wl']-0.5*item['delta_wl']
    item['end_wl'] = item['mean_wl']+0.5*item['delta_wl']
    item['span_10pts'] = _gen_lam_for_filter(item)


def deproj_image(x, y, inc, pa, polar_in=True, polar_out=True):
    return _deproj(x=x, y=y, inc=inc, pa=pa, polar_in=polar_in,
                   polar_out=polar_out, fourier_plan=False)


def deproj_fourier(x, y, inc, pa, polar_in=True, polar_out=True):
    return _deproj(x=x, y=y, inc=inc, pa=pa, polar_in=polar_in,
                   polar_out=polar_out, fourier_plan=True)


def _deproj(x, y, inc, pa, polar_out=True, polar_in=True, fourier_plan=False):
    if polar_in:
        y, x = x*np.cos(y), x*np.sin(y)
    else:
        y, x = x, y
    if fourier_plan:
        X = (x*np.cos(pa) + y*np.sin(pa))*np.cos(inc)
        Y = y*np.cos(pa) - x*np.sin(pa)
    else:
        X = x*np.cos(pa) + y*np.sin(pa)
        Y = (y*np.cos(pa) - x*np.sin(pa))/np.cos(inc)
    x, y = np.hypot(Y, X), (np.arctan2(X, Y)-pa) % (2*np.pi)
    if not polar_out:
        return x*np.cos(y), x*np.sin(y)
    return x, y


def ratio_bb_flux(wl, teff1, teff2, diam1, diam2):
    """
    Calculates the flux ratio at wl (meter) of 2 black-bodies of
    temperature teff (K) and diameter diam
    """
    # pi*2*h*c**2 = 3.741771524664128e-16
    # h*c/kb = 0.014387769599838155
    res = np.exp(0.014387769599838155/(wl*teff2))-1
    res /= np.exp(0.014387769599838155/(wl*teff1))-1
    res *= (diam1/diam2)**2
    return res


def mag2diam(ref_mag, ref_band, teff, nbr_pts=10):
    """
    Returns the DIAMETER (not radius) of a black body of a given
    magnitude (U,B,V,R,I,J,H,K)
    """
    filt = ALL_FILTERS[ref_band]
    if nbr_pts != 10:
        lam = _gen_lam_for_filter(filt, nbr_pts)
    else:
        lam = filt['span_10pts']

    # flux in watts from magnitude
    res = filt['flux_wm2m']*10**(-0.39809*ref_mag)*filt['delta_wl']
    # integrated flux
    res /= np.trapz(blackbody_spectral_irr(teff, lam), lam)
    return 2*RAD2MAS*np.sqrt(res)


def blackbody_spectral_irr(teff, wl):
    """
    Calculates the emitted (at surface!) flux in W/m2/m assuming black
    body behaviour, given an effective temperature (K) and wl (meter)
    """
    # pi*2*h*c**2 = 3.741771524664128e-16
    # h*c/kb = 0.014387769599838155
    res = 3.741771524664128e-16/wl**5
    res /= np.exp(0.014387769599838155/(wl*teff))-1
    return res


def hduToDataType(hdu):
    """
    Give a hdu, return a datatype name as string or None if it is not
    a data hdu
    """
    hdu = getattr(hdu, "header", {})
    return DATATYPEEXTNAMES.get(hdu.get('EXTNAME', None), None)


def hduWlindex(hdus, returnHdu=False):
    """
    Give a list of hdus, return the index of the OI_WAVELENGTH hdu, or
    False if not found
    """
    for idx, hdu in enumerate(hdus):
        if getattr(hdu, "header", {}).get('EXTNAME', None) == 'OI_WAVELENGTH':
            if returnHdu:
                return idx, hdu
            else:
                return idx
    else:
        return False


def hduToColNames(hdu):
    """
    Returns a list of the column names
    """
    return [i.name for i in hdu.data.columns.columns]


def replicate(ar, dims):
    left = ()
    right = ()
    hasbeennone = False
    for i in dims:
        if i is None or i == -1:
            hasbeennone = True
            continue
        if hasbeennone:
            right += (i,)
        else:
            left += (i,)
    ar = np.tile(ar, left+tuple(np.ones(ar.ndim).astype(int)))
    if right != ():
        ar = ar.repeat(np.prod(right)).reshape(ar.shape+right)
    return ar


class font:
    white = '\033[97m'
    black = '\033[38;5;16m'
    gray = '\033[90m'
    red = '\033[31m'
    green = '\033[32m'
    yellow = '\033[33m'
    orange = '\033[38;5;166m'
    blue = '\033[34m'
    magenta = '\033[35m'
    nocolor = '\033[39m'
    bold = '\033[1m'
    nobold = '\033[21m'
    underlined = '\033[4m'
    nounderlined = '\033[24m'
    dim = '\033[2m'
    nodim = '\033[22m'
    normal = nodim + nobold + nobold + nocolor
    clear = chr(27)+"[2J"

    def pos(self, line, col):
        return "\033[{};{}H".format(int(line), int(col))


def astroskyplot(radius, polar=False, unit='arcsec', sameaxes=True):
    if polar:
        fig, ax = matplotlibpyplotsubplots(subplot_kw={'projection': 'polar'})
        ax.set_theta_zero_location('N')
        ax.set_ylim([0, float(radius)])
        ax.set_xticklabels(['N', u'45°', 'E', u'135°',
                            'S', u'225°', 'W', u'315°'])
        ax.set_xlabel('[{}]'.format(unit))
    else:
        fig, ax = matplotlibpyplotsubplots()
        ax.set_xlim([-float(radius), float(radius)])
        ax.set_ylim([-float(radius), float(radius)])
        ax.invert_xaxis()
        ax.set_ylabel('E, dec [{}]'.format(unit))
        ax.set_xlabel('S, ra [{}]'.format(unit))
    if sameaxes:
        ax.set_aspect('equal')
    return fig, ax


def gen_seed():
    res = int(str((timetime() % 3600)/3600)[2:])
    return int(res*multiprocessingcurrent_process().pid % sysmaxint)


def gen_generator(seed=None):
    if seed is None:
        return np.random.RandomState(gen_seed())
    return np.random.RandomState(seed)


def random_custom_pdf(x, pdf, size=1, renorm=False, seed=None):
    # if any negative value in pdf, raises error
    if (np.asarray(pdf) < 0).any():
        raise Exception("Can't compute pdf with negative values")
    cdf = integrate_array(x, pdf)
    return random_custom_cdf(x, cdf, size=size, renorm=renorm, seed=seed)


def random_custom_cdf_fct(x, cdf, renorm=False):
    """
    """
    cdf = np.asarray(cdf)
    if (np.diff(cdf) < 0).any():
        raise Exception("cdf must be monotonic increasing")
    if not renorm:
        # if first or last element not close enough from 0 and 1, rejects
        if np.round(cdf[-1], 10) != 1 or np.round(cdf[0], 10) != 0:
            raise Exception("Wrong cdf distribution (first element not 0 \
                                or last element not 1). Use 'renorm'")
    # forces first and last elements to be 0 and 1 exactly
    cdf = (cdf-cdf[0])/(cdf[-1]-cdf[0])
    return scipyinterpolateinterp1d(cdf, x, kind='linear')


def random_custom_cdf(x, cdf, size=None, renorm=False, seed=None):
    """
    if size is False or None: returns a function
    if size == 1 : returns a float
    if size
    """
    fct = random_custom_cdf_fct(x=x, cdf=cdf, renorm=renorm)
    if size is None or size is False:  # defaut
        return fct
    rnd = gen_generator(seed=seed)
    if size == 1:
        return float(fct(rnd.uniform(size=1)))
    elif np.size(size) == 1:
        return fct(rnd.uniform(size=size))
    else:
        return fct(rnd.uniform(size=np.prod(size)).reshape(size))


def integrate_array(x, y, axis=0):
    y = np.array(y).swapaxes(axis, -1)
    inty = np.zeros(y.shape)
    myslice = [slice(0, i)
               for i in aslist(y.shape)[:-1]]+[slice(1, y.shape[-1])]
    inty[myslice] = scipyintegratecumtrapz(y, x)
    return inty.swapaxes(axis, -1)


def aslist(data, numpy=False, integer=False):
    if not hasattr(data, "__iter__") or isinstance(data, dict):
        if integer:
            ret = np.asarray([int(data)])
        else:
            ret = np.asarray([data])
    else:
        if integer:
            ret = np.asarray(data).flatten().astype(int)
        else:
            ret = np.asarray(data).flatten()
    if not numpy:
        return list(ret)
    return ret


def clean_name(txt):
    authorized = range(65, 91)+range(48, 58)+range(97, 123)
    return "".join([letter if (ord(letter) in authorized) else ""
                    for letter in str(txt)])


def colorbar(cmap="jet", cm_min=0, cm_max=1):
    if isinstance(cmap, str):
        cmap = cmget_cmap(cmap)
    norm = matplotlibpyplotNormalize(cm_min, cm_max)
    mappable = cmScalarMappable(cmap=cmap, norm=norm)
    mappable._A = []
    return cmap, norm, mappable


def ident(ar, *args, **kwargs):
    return ar


def _vv(img, masperpx, u, v, wl):
    demifov = 0.5*np.asarray(img.shape)*masperpx
    Y = np.linspace(demifov[0], -demifov[0], img.shape[0])
    X = np.linspace(demifov[1], -demifov[1], img.shape[1])
    Y, X = np.meshgrid(Y, X)
    X = X.reshape(X.shape+(1,))
    X = X.astype(np.float32)
    dum = X*u
    X = None
    Y = Y.reshape(Y.shape+(1,))
    Y = Y.astype(np.float32)
    dum += Y*v
    dum *= DEUXPI*MAS2RAD/wl
    return dum.astype(np.complex64)


def calcImgVis(img, masperpx, u, v, wl):
    """
    img: the image, (x, y) dimension
    masperpx the mas/px values for the images (x, y) in mas/px
    u, v: the coordinates on which to calculate the FT
    wl: the wavelength
    """
    dum = _vv(img, masperpx, u, v, wl)
    dum *= 1j
    dum = np.exp(dum)
    dum = dum.T
    dum *= img
    dum = dum.T
    if OLDNUMPY:
        return dum.sum(axis=0).sum(axis=0)
    else:
        return dum.sum(axis=(0, 1))


def getDetails(hdu):
    """
    Give a hdu
    Returns (ndata, nset, nunique, nholes, nwl)
    """
    allstation = hdu.data['STA_INDEX']
    ndata = allstation.shape[0]
    t3unique = 1.
    if allstation.shape[1] == 3:
        t3unique = np.exp(1./allstation[:, 2])
    nunique = len(set((allstation[:, 0] + 1j*allstation[:, 1])*t3unique))
    if allstation.shape[1] == 3:  # T3
        nholes = NHOLES[np.argmin(np.abs(TRIHOLESTAB-nunique))]
    else:
        nholes = NHOLES[np.argmin(np.abs(BLHOLESTAB-nunique))]
    nset = ndata/nunique
    if hdu.data[DATAEXTNAMES[hdu.header.get('EXTNAME')]].ndim == 1:
        nwl = 1
    else:
        nwl = hdu.data[DATAEXTNAMES[hdu.header.get('EXTNAME')]].shape[1]
    return ndata, nset, nunique, nholes, nwl


def gethduMJD(hdu, withdet=False):
    """
    Returns, taken from the hdu, the unsorted mjd (MJD) and target
    sorted-index related to the MJD (targetsortmjd), such that
    MJD[targetsortmjd] gives sorted MJD values.
    In case several datasets have the same mjd value, this function
    uses the intefration time to infer mjd values which vary from
    dataset to dataset.
    Output shape = (nset, nunique) for MJD; (nset) for targetsortmjd
    """
    ndata, nset, nunique, nholes, nwl = getDetails(hdu)
    mjd = hdu.data['MJD'].reshape((nset, nunique))
    # the mjd information is sometimes the same = bullshit
    if np.allclose(mjd.std(axis=0), 0):
        # "fake" fix it using integration time
        delta = (hdu.data['INT_TIME'][np.arange(0, ndata, nunique)]*nunique)
        mjd += replicate(np.clip(delta, 30., 1e9).cumsum()/SECPERDAY,
                              (-1, nunique))
    sortmjd = np.argsort(mjd[:,0])
    if withdet:
        return sortmjd, mjd, (ndata, nset, nunique, nholes, nwl)
    else:
        return sortmjd, mjd


def unique(ar, precision, return_index=False, return_inverse=False,
           return_counts=False):
    """
    Returns the unique numbers of an array, at a certain precision

    Refer to np.unique
    """
    return np.unique(np.floor(ar/precision).astype(int)*precision,
                     return_index=return_index,
                     return_inverse=return_inverse,
                     return_counts=return_counts)


def maskedshape(masked_shape, size_unmasked):
    totsize = np.prod(masked_shape) + size_unmasked
    percentmasked = np.round(size_unmasked*1./totsize, 3)*100
    return "{} - {}% masked".format(masked_shape, percentmasked)


def round_fig(x, n=1, retint=False):
    """
    Rounds x at the n-th figure. n must be >1
    ex: 1234.567 with n=3-> 1230.0
    """
    if np.iterable(x):
        x = np.asarray(x).copy()
        ff = (x != 0)
        dd = 10**(np.floor(np.log10(np.abs(x[ff])))-n+1)
        x[ff] = np.round(x[ff]/dd)
        if not retint:
            x[ff] *= dd
        return x
    elif x != 0:
        dd = 10**(np.floor(np.log10(np.abs(x)))-n+1)
        x = np.round(x/dd)
        if not retint:
            x *= dd
        return x
    else:
        return x


def quantile(x, q, weights=None):
    """
    Like numpy.percentile, but:
    * Values of q are quantiles [0., 1.] rather than percentiles [0., 100.]
    * scalar q not supported (q must be iterable)
    * optional weights on x
    """
    if weights is None:
        return np.percentile(x, [100. * qi for qi in q])
    else:
        idx = np.argsort(x)
        xsorted = x[idx]
        cdf = np.add.accumulate(weights[idx])
        cdf /= cdf[-1]
        return np.interp(q, cdf, xsorted).tolist()


def psf(lOverd, masperpx):
    """
    lOverd in radian
    masperpx in mas per px
    """
    nbpts = (lOverd*4/(masperpx*MAS2RAD))//2*2+1
    y, x = np.meshgrid(np.linspace(-1, 1, nbpts), np.linspace(-1, 1, nbpts))
    psf = airy(np.hypot(y, x)*2*lOverd+1e-10, 1/lOverd)**2
    psf /= psf.sum()
    return psf


def airy(th, B, lam=None):
    """
    Return the visibility value (unsquared), given
    - th the angular diameter in radian
    - B the baseline in m (alternatively in m/lambda)
    - lam the wavelength (alternatively None if B is already given as m/lambda)
    """
    if lam is None:
        x = np.pi*th*B
    else:
        x = np.pi*th*B/lam
    return 2*airyJ1(x)/x


def gauss2D(x, y, A, x0, y0, sigmaX, sigmaY, theta):
    sx2 = 1./sigmaX**2
    sy2 = 1./sigmaY**2
    st2 = np.sin(theta)**2
    ct2 = np.cos(theta)**2
    s2t = np.sin(2*theta)
    a = ct2*sx2 + st2*sy2
    b = s2t*(sy2 - sx2)
    c = st2*sx2 + ct2*sy2
    # a = (np.cos(theta)/sigmaX)**2 + (np.sin(theta)/sigmaY)**2
    # b = np.sin(2*theta)/sigmaY**2 - np.sin(2*theta)/sigmaX**2
    # c = (np.sin(theta)/sigmaX)**2 + (np.cos(theta)/sigmaY)**2
    return A*np.exp(-0.5*(a*(x-x0)**2 + b*(x-x0)*(y-y0) + c*(y-y0)**2))


def gauss1D(x, y, A, x0, y0, sigma):
    return A*np.exp(-0.5*((x-x0)**2 + (y-y0)**2)/sigma**2)


def gauss1Dsimple(blwlsigma):
    return np.exp(-blwlsigma*blwlsigma)


class PriorWrapper(object):
    def __init__(self, fct, makenorm=True, makeunlog=False, fct_log=False,
                 npts=1000, *args, **kwargs):
        if not callable(fct):
            raise Exception("Not callable")
        self.fct = fct
        self.args = args
        self.kwargs = kwargs
        # apply exp on a ln-function at callback or not
        self.makeunlog = makeunlog is True
        # renorm at callback or not
        self.makenorm = makenorm is True
        # whether to add or divide the normalization
        self.fct_log = fct_log is not False and makeunlog is False
        self.prior_log = kwargs.get('prior_log', False)  # jeffrey
        if kwargs.get('prior_bounds', None) is None:
            raise Exception("No prior bounds found")
        if self.prior_log:
            self.x = np.logspace(np.log(kwargs.get('prior_bounds')[0]),
                                 np.log(kwargs.get('prior_bounds')[1]),
                                 npts,
                                 base=np.e)
        else:
            self.x = np.linspace(kwargs.get('prior_bounds')[0],
                                 kwargs.get('prior_bounds')[1],
                                 npts)
        if makenorm is True:
            resfct = self.fct(self.x, *self.args, **self.kwargs)
            if self.fct_log:
                self.normval = np.log(np.trapz(np.exp(resfct), self.x))
            elif self.makeunlog:
                self.normval = np.trapz(np.exp(resfct), self.x)
            else:
                self.normval = np.trapz(resfct, self.x)

    def __call__(self, x, *args, **kwargs):
        mergeargs = args + self.args[len(args):]
        mergekwargs = {}
        mergekwargs.update(self.kwargs, **kwargs)
        ret = self.fct(x, *mergeargs, **mergekwargs)
        if self.makeunlog:
            ret = np.exp(ret)
        if self.makenorm:
            if self.fct_log:
                ret -= self.normval
            else:
                ret /= self.normval
        return ret



#!/usr/bin/env python
# -*- coding: utf-8 -*-

###############################################################################
#
#  SOIF - Sofware for Optical Interferometry fitting
#  Copyright (C) 2016  Guillaume Schworer
#
#  This program is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
#  For any information, bug report, idea, donation, hug, beer, please contact
#    guillaume.schworer@obspm.fr
#
###############################################################################


import matplotlib.pyplot as plt
from matplotlib.gridspec import GridSpec as _matplotlibGridspecGridSpec
from matplotlib.patches import FancyArrowPatch as _matPatFancyArrowPatch
from copy import deepcopy
from time import strftime
try:
    import astropy.io.fits as pf
except ImportError:
    import pyfits as pf

from .oiunitmodels import *
from .oimainobject import Oimainobject
from . import oiexception as exc
from . import core
from .oifits import Oifits
np = core.np

__all__ = ['Oimodel']


class Oimodel(object):
    def __init__(self, oidata=None, objs=[], tweakparams=None, **kwargs):
        self._objs = []
        self.raiseError = bool(kwargs.pop('raiseError', True))
        if isinstance(oidata, Oifits):
            self.oidata = oidata
        else:
            self.oidata = None
        if tweakparams is not None and not callable(tweakparams):
            if exc.raiseIt(exc.NotCallable, self.raiseError, fct="tweakparams"):
                return False
        self._tweakparams = tweakparams
        self.nobj = 0
        self._nparamsObj = 0
        if not hasattr(objs, "__iter__"):
            self.add_obj(objs)
        else:
            for item in objs:
                self.add_obj(item)

    def _info(self):
        txt = core.font.blue+"<SOIF Model>{}\n {} objects:\n  {}".format(
                            core.font.normal,
                            self.nobj,
                            "\n  ".join(map(str, self._objs)))
        if self._hasdata:
            return "{}\n{}".format(txt, str(self.oidata))
        else:
            return txt

    def __repr__(self):
        return self._info()

    __str__ = __repr__

    @property
    def _hasdata(self):
        return self.oidata is not None

    @property
    def nparams(self):
        if self._hasdata:
            return self.nparamsObjs + int(self.oidata.systematic_fit)
        else:
            return self.nparamsObjs

    @nparams.setter
    def nparams(self, value):
        exc.raiseIt(exc.ReadOnly, self.raiseError, attr="nparams")

    @property
    def nparamsObjs(self):
        self._nparamsObj = 0
        for item in self._objs:
            self._nparamsObj += item._nparams
        return self._nparamsObj

    @nparamsObjs.setter
    def nparamsObjs(self, value):
        exc.raiseIt(exc.ReadOnly, self.raiseError, attr="nparamsObjs")

    def add_obj(self, typ, name=None, params={}, prior={}):
        """
        Adds an object to the model
        """
        # gets the future new name of the object
        if isinstance(typ, Oimainobject):
            name = typ.name
        else:
            name = core.clean_name(name)
        # check for already existing name
        if hasattr(self, "o_"+name):
            if exc.raiseIt(exc.BusyName, self.raiseError, name=name):
                return
        # if all good, proceeds
        if isinstance(typ, Oimainobject):  # if first argument is already a built object
            self._objs.append(deepcopy(typ))
        elif typ in _objects:  # if we need to build the object, and it exists
            self._objs.append(globals()[typ](name=name, params=params, prior=prior))
        else:
            if exc.raiseIt(exc.InalidUnitaryModel, self.raiseError, typ=typ):
                return
        setattr(self, "o_"+name, self._objs[-1])  # quick access as a class property
        if hasattr(self._objs[-1], 'prepare') and self._hasdata:
            self._objs[-1].prepare(oidata=self.oidata)
        self.nobj += 1
        # dum = self.nparamsObjs

    def del_obj(self, idobj):
        """
        Delete an object from the model.
        idobj can be the name of the object or its index in the model
        list.
        """
        if not isinstance(idobj, int):
            name = core.clean_name(idobj)
            if not hasattr(self, "o_"+name):
                if exc.raiseIt(exc.NotFoundName, self.raiseError, name=name):
                    return
            ind = self._objs.index(getattr(self, "o_"+name))
        else:
            ind = idobj
            name = self._objs[ind].name
        self._objs.pop(ind)
        delattr(self, "o_"+name)
        print("{}Deleted object '{}'{}.".format(core.font.blue, name, core.font.normal))
        self.nobj -= 1
        dum = self.nparamsObjs

    def getP0(self):
        """
        Return an initialized param list
        """
        ret = []
        for item in self._objs:
            ret += item.getP0()
        if getattr(self.oidata, "systematic_fit", False):
            ret += [self.oidata.systematic_p0()]
        return ret


    @property
    def paramstr(self):
        ret = []
        for item in self._objs:
            for arg in item._pkeys:
                ret.append(item.name+"_"+arg)
        if getattr(self.oidata, "systematic_fit", False):
            ret += ["sys"]
        return ret
    @paramstr.setter
    def paramstr(self, value):
        exc.raiseIt(exc.ReadOnly, self.raiseError, attr="paramstr")


    @property
    def params(self):
        """
        Return current params
        """
        ret = []
        for item in self._objs:
            ret += getattr(item, "params", [])
        if getattr(self.oidata, "systematic_fit", False):
            fiterrors = self.oidata.systematic_prior is not None
            ret += [self.oidata.systematic_prior if fiterrors else self.oidata.systematic_p0()]
        return ret
    @params.setter
    def params(self, value):
        self.setParams(value)


    def reinit(self, params):
        self.setParams(params=params, priors=True)


    def statut(self, params, customlike=None, **kwargs):
        self.setParams(params)
        chi2 = self.likelihood(params=params, customlike=customlike, chi2=True, **kwargs)
        print('Khi-2: {:.3f}'.format(chi2))
        print('Khi-2 NULL: {:.3f}'
            .format(self.likelihood(params=params, customlike=customlike, chi2=True, null=True, **kwargs)))
        print('ln-like: {:.3f}'.format(self.likelihood(params=params, customlike=customlike, chi2=False, **kwargs)))
        print('ln-like NULL: {:.3f}'
            .format(self.likelihood(params=params, customlike=customlike, chi2=False, null=True, **kwargs)))
        print('Number of parameters (k): {:d}'.format(self._nparamsObj))


    def setParams(self, params, priors=False):
        if len(params) != self.nparams:
            if exc.raiseIt(exc.BadParamsSize, self.raiseError, size=self.nparams):
                return
        parampos = 0
        for item in self._objs:
            item.setParams(params=params[parampos:parampos+item._nparams], priors=priors)
            parampos += item._nparams
        if getattr(self.oidata, "systematic_fit", False):
            self.oidata.systematic_prior = params[parampos]
            self.oidata._systematic_prior = self.oidata.systematic_prior


    def compVis(self, params=None, u=None, v=None, wl=None):
        """
        Calculates the complex visibility of the model from all unitary models
        """
        if self._hasdata:
            totviscomp = self._compVis(u=self.oidata.uvwl['u'], v=self.oidata.uvwl['v'], wl=self.oidata.uvwl['wl'],
                blwl=self.oidata.uvwl['blwl'], params=params)
            return self.oidata.remorph(totviscomp)
        else:
            if u is None or v is None or wl is None:
                if exc.raiseIt(exc.NoDataModel, self.raiseError):
                    return
            else:
                return self._compVis(u=u, v=v, wl=wl, params=params)


    def calcuvimage(self, blmax, wl, params=None, nbpts=101):
        """
        Outputs the complex visibility for a grid a (u,v) in [-blmax,blmax], with nbpts in each dimension
        """
        u, v = np.meshgrid(np.linspace(-blmax, blmax, nbpts), np.linspace(-blmax, blmax, nbpts))
        return self._compVis(u=u, v=v, wl=wl, blwl=np.hypot(u, v)/wl, params=params)


    def _compVis(self, u, v, wl, blwl, params=None):
        parampos = 0
        totFluxvis = 0.
        totviscomp = np.zeros(u.shape, dtype=complex) # initialize array
        if params is None:
            params = self.getP0() # initialize at p0 in case no params is given
        else:
            self.setParams(params)
        if self._tweakparams is not None:
            self._tweakparams(self, params)
        for item in self._objs:
            compvis, flx = item._calcCompVis(u=u, v=v, wl=wl, blwl=blwl)
            compvis *= flx
            totviscomp += compvis
            totFluxvis += flx
            parampos += item._nparams
        totviscomp /= totFluxvis
        return totviscomp


    def calcimage(self, params=None, sepmax=None, wl=None, masperpx=None, nbpts=101, psfConvolve=None, **kwargs):
        """
        psfConvolve in mas (lambda/D)
        """
        parampos = 0
        totFluxvis = 0.
        # check for set-resolution objects
        masperpxfixed = None
        nbptscheck = nbpts
        if nbpts is None: nbptscheck = int(2.*sepmax/masperpx)
        for item in self._objs:
            if hasattr(item, "_masperpx"):
                if masperpxfixed is None:
                    masperpxfixed = item._masperpx
                elif np.abs(1-masperpxfixed/item._masperpx)*nbptscheck > 1:
                    if exc.raiseIt(exc.MasperpxMismatch, self.raiseError, mpp1=masperpxfixed, mpp2=item._masperpx):
                        return
        if masperpxfixed is not None:
            masperpx = masperpxfixed
        # check resolution parameters
        if sepmax is None and masperpx is None and nbpts is not None:
            sepmax = 1000*nbpts/(2*np.hypot(self.oidata.uvwl['v'], self.oidata.uvwl['u']).max())
        if masperpx is not None and sepmax is not None:
            nbpts = int(2.*sepmax/masperpx)
        elif masperpx is not None and nbpts is not None:
            pass
        else:
            masperpx = sepmax*2./nbpts
        #elif sepmax is not None and nbpts is not None:
        #    masperpx = sepmax*2./nbpts
        #else:
        #    raise Exception("can't determine the resolution, nbpts should not be None")
        sepmax = 0.5*nbpts*masperpx
        # check other parameters
        if wl is None: wl = self.oidata.uvwl['wl'].min()
        # initialize array
        img = np.zeros((nbpts, nbpts))
        if self._tweakparams is not None: self._tweakparams(self, params)
        if params is None:
            params = self.getP0() # initialize at p0 in case no params is given
        else:
            self.setParams(params)
        for item in self._objs:
            img += item.image(sepmax=sepmax, masperpx=masperpx, wl=wl, nbpts=nbpts)
            parampos += item._nparams
        if psfConvolve is not None:
            if len(core.aslist(psfConvolve)) == 1:
                psf = core.psf(float(psfConvolve)*core.MAS2RAD, masperpx)
            elif len(core.aslist(psfConvolve)) == 3:
                sx, sy, th = map(float, psfConvolve[:3])
                y, x = np.meshgrid(np.arange(nbpts)-nbpts//2, np.arange(nbpts)-nbpts//2)
                psf = core.gauss2D(x, y, 1, 0, 0, sx*0.5/masperpx, sy*0.5/masperpx, th)
            img = core.fftconvolve(img, psf, mode='same')
        if kwargs.pop('retresol', False):
            return img, (sepmax, masperpx, nbpts)
        return img


    def image(self, params=None, sepmax=None, wl=None, masperpx=None, nbpts=101, cmap='jet', cm_min=None,
              cm_max=None, ret=False, visu=None, psfConvolve=None, **visuargs):
        """
        psfConvolve in mas (lambda/D)
        """
        # check other params
        if visu is None:
            visu = core.ident
        if not callable(visu):
            if exc.raiseIt(exc.NotCallable, self.raiseError, fct='visu'):
                return
        toplot, (sepmax, masperpx, nbpts) = self.calcimage(sepmax=sepmax, wl=wl, params=params, nbpts=nbpts, psfConvolve=psfConvolve, retresol=True)
        toplot = visu(toplot, **visuargs)
        if cm_min is None:
            cm_min = toplot.min()
        if cm_max is None:
            cm_max = toplot.max()
        cmap, norm, mappable = core.colorbar(cmap=cmap, cm_min=cm_min, cm_max=cm_max)
        thefig, ax = core.astroskyplot(sepmax, polar=False, unit='mas')
        ax.matshow(toplot, origin='lower', extent=[-sepmax, sepmax, -sepmax, sepmax], cmap=cmap, norm=norm)
        plt.colorbar(mappable)
        if ret:
            return toplot


    def uvimage(self, params=None, blmax=None, wl=None, typ='vis2', nbpts=101, cmap='jet', cm_min=None,
                cm_max=None, ret=False, visu=None, **visuargs):
        """
        typ can be: 'VIS2', 'T3PHI', 'T3AMP', 'VISPHI', or 'VISAMP'
        """
        # check other params
        if visu is None: visu = core.ident
        if not callable(visu):
            if exc.raiseIt(exc.NotCallable, self.raiseError, fct='visu'):
                return
        toplot = self.calcuvimage(blmax=blmax, wl=wl, params=params, nbpts=nbpts)
        toplot = visu(core.FCTVISCOMP.get(typ.upper(), core.FCTVISCOMP['VIS2'])(toplot), **visuargs)
        if cm_min is None:
            cm_min = toplot.min()
        if cm_max is None:
            cm_max = toplot.max()
        cmap, norm, mappable = core.colorbar(cmap=cmap, cm_min=cm_min, cm_max=cm_max)
        thefig, ax = core.astroskyplot(blmax, polar=False, unit='m')
        ax.matshow(toplot, origin='lower', extent=[-blmax, blmax, -blmax, blmax], cmap=cmap, norm=norm)
        plt.colorbar(mappable)
        if ret:
            return toplot

    def residual(self, params, c=None, cmap='jet', cm_min=None, cm_max=None, datatype='All'):
        if not self._hasdata:
            if exc.raiseIt(exc.NoDataModel, self.raiseError): return
        calcindex = {'vis2':0, 't3phi':1, 't3amp':2, 'visphi':3, 'visamp':4}
        fullmodel = self.compVis(params=params)
        cm_min_orig = cm_min
        cm_max_orig = cm_max
        if datatype.lower() == 'all':
            datatype = core.DATAKEYSLOWER
        elif np.iterable(datatype)==1 and not isinstance(datatype, str):
            datatype = [str(dum).lower() for dum in datatype if dum.lower() in core.DATAKEYSLOWER]
        else:
            datatype = [str(dum).lower() for dum in [str(datatype)] if dum.lower() in core.DATAKEYSLOWER]
        if self.oidata.systematic_fit: self.oidata.systematic_prior = params[self._nparamsObj]
        for index, datatypeloop in enumerate(datatype):
            if not getattr(self.oidata, datatypeloop): continue
            data = getattr(self.oidata, datatypeloop).data
            model = fullmodel[calcindex[datatypeloop]]
            error = getattr(self.oidata, datatypeloop).error
            if self.oidata.systematic_fit: error = np.sqrt(error*error + self.oidata.systematic_prior*self.oidata.systematic_prior)
            thefig = plt.figure()
            if c is not None:
                grid = _matplotlibGridspecGridSpec(5, 21)
            else:
                grid = _matplotlibGridspecGridSpec(5, 20)
            theax = thefig.add_subplot(grid[0:4,0:20])
            theotherax = thefig.add_subplot(grid[4,0:20], sharex=theax)
            for k, v in np.ndenumerate(data):
                theax.add_patch(_matPatFancyArrowPatch((model[k], data[k]),(model[k], data[k]+error[k]), arrowstyle='-['))
                theax.add_patch(_matPatFancyArrowPatch((model[k], data[k]),(model[k], data[k]-error[k]), arrowstyle='-['))
            mini = min(data.min(), model.min())
            maxi = max(data.max(), model.max())
            theax.plot([mini, maxi], [mini, maxi], 'g-', lw=2)
            if c is not None:
                thecb = thefig.add_subplot(grid[:,20])
                thecb.set_xticklabels('')
                thecb.yaxis.tick_right()
                colorkeyattr = str(c)
                if not hasattr(getattr(self.oidata, datatypeloop), colorkeyattr): colorkeyattr = 'wl'
                colorattr = getattr(getattr(self.oidata, datatypeloop), colorkeyattr)
                #if datatypeloop[:2]=='t3':
                #    if colorattr.shape != self.oidata._t3outputshape:
                #        colorattr = colorattr.mean(axis=-1)
                if cm_min_orig is None: cm_min = colorattr.min()
                if cm_max_orig is None: cm_max = colorattr.max()
                cmap, norm, mappable = core.colorbar(cmap=cmap, cm_min=cm_min, cm_max=cm_max)
                theax.scatter(x=model, y=data, s=30, c=colorattr, cmap=cmap, norm=norm, marker='o', edgecolors='none', alpha=1)
                div = np.true_divide(model-data, error)
                theotherax.scatter(x=model, y=self.oidata.erb_sigma(np.abs(div))*np.sign(div), s=30, c=colorattr, cmap=cmap, norm=norm, marker='o', edgecolors='none', alpha=1)
                thefig.colorbar(mappable=mappable, cax=thecb)
                colortitle = ', color= '+colorkeyattr
            else:
                theax.scatter(x=model, y=data, s=30, marker='o', edgecolors='none', alpha=1)
                div = np.true_divide(model-data, error)
                theotherax.scatter(x=model, y=self.oidata.erb_sigma(np.abs(div))*np.sign(div), s=30, marker='o', edgecolors='none', alpha=1)
                colortitle = ''
            theotherax.grid(True)
            theax.set_title(datatypeloop+colortitle)


    #def grid(self, param_x, param_y, bin_x=50, bin_y=50, mode=np.max, cmap="jet"):
    #    pass

    def likelihood(self, params, customlike=None, chi2=False, **kwargs):
        if not self._hasdata:
            if exc.raiseIt(exc.NoDataModel, self.raiseError): return
        kwargs['chi2'] = chi2
        return standardLikelihood(params=params, model=self, customlike=customlike, kwargs=kwargs)


    def save(self, filename, clobber=False):
        hdulist = pf.HDUList()
        ext = '.oif.fits'
        if filename.find(ext)==-1: filename += ext
        hdu = pf.PrimaryHDU()
        hdu.header.set('EXT', 'MODEL', comment='Type of information in the HDU')
        hdu.header.set('DATE', strftime('%Y%m%dT%H%M%S'), comment='Creation Date')
        hdu.header.set('NOBJ', self.nobj, comment='Number of objects in the model')
        hdu.header.set('NPARAMS', self.nparams, comment='Total number of free parameters')
        for i, item in enumerate(self._objs):
            hdu.header.set('OBJ'+str(i), item.name, comment='Name of object '+str(i))
        allmodes = ['vis2', 't3phi', 't3amp', 'visphi', 'visamp']
        for mode in allmodes:
            number = []
            if getattr(self.oidata, mode):
                number = [len(i) for i in getattr(self.oidata, "_input_"+mode) if i is not None]
            hdu.header.set('N'+mode, int(np.sum(number)), comment='Total number of '+mode+' measurements')
            hdu.header.set('F_'+mode, len(number), comment='Number of '+mode+' files')

        hdu.header.add_comment('Written by Guillaume SCHWORER')
        hdulist.append(hdu)

        hdulist.writeto(filename, clobber=clobber)

        for item in self._objs:
            item.save(filename, append=True, clobber=clobber)

        if self._hasdata: self.oidata.save(filename, append=True, clobber=clobber)

        return filename


def standardLikelihood(params, model, customlike=None, kwargs={}):
    '''
    Calculate the likelihood of the model versus the data.
    Use customlike and kwargs parameters for a custom likelihood function.
    This customlike function must accept input params:
    - params (list of the p values for that step)
    - modeledData (tuple of model vis2, t3phi, t3amp, visphi, visamp given the p value of the parameters). The uncalculated data is set to None.
    - model (oifiting object)

    And must return the log-likelihood of that step
    '''
    null = kwargs.pop('null', False)
    chi2 = kwargs.pop('chi2', False)
    quality = 0.
    ln_prior = 0.
    if model.oidata.systematic_fit:
        model.oidata.systematic_prior = params[model._nparamsObj]
    if null: # if we just want NULL hypothesis
        vis2, t3phi, t3amp, visphi, visamp = 1., 0., 1., 0., 1.
    else:
        parampos = 0
        for item in model._objs:
            for arg in item._pkeys:
                pb = getattr(item, arg+'_bounds')
                if not (pb[0] <= params[parampos] <= pb[1]):
                    return -np.inf # exit with -inf if params outside bounds
                parampos += 1
        if model.oidata.systematic_fit:
            if not (model.oidata.systematic_bounds[0] <= model.oidata.systematic_prior <= model.oidata.systematic_bounds[1]):
                return -np.inf # exit with -inf if params outside bounds
            #parampos += 1
        if customlike is not None: return customlike(params=params, model=model, **kwargs)
        if not chi2:
            pass
            #ln_prior += getattr(item, arg+'_prior_lnfunc')(params[parampos], params=params, parampos=parampos, **getattr(item, arg+'_prior_kwargs'))
        if model.nobj==0:
            vis2, t3phi, t3amp, visphi, visamp = 1., 0., 1., 0., 1.
        else:
            vis2, t3phi, t3amp, visphi, visamp = model.compVis(params=params)
    if model.oidata.vis2:
        if model.oidata.systematic_fit:
            invvar = 1./(model.oidata.vis2.error*model.oidata.vis2.error + model.oidata.systematic_prior*model.oidata.systematic_prior)
        else:
            invvar = model.oidata.vis2._invvar
        if chi2:
            quality += ((model.oidata.vis2.data - vis2)**2*invvar).sum()
        else:
            quality += ((model.oidata.vis2.data - vis2)**2*invvar - np.log(invvar)).sum()
    if model.oidata.visphi:
        if model.oidata.systematic_fit:
            invvar = 1./(model.oidata.visphi.error*model.oidata.visphi.error + model.oidata.systematic_prior*model.oidata.systematic_prior)
        else:
            invvar = model.oidata.visphi._invvar
        if chi2:
            quality += ((model.oidata.visphi.data - visphi)**2*invvar).sum()
        else:
            quality += ((model.oidata.visphi.data - visphi)**2*invvar - np.log(invvar)).sum()
    if model.oidata.visamp:
        if model.oidata.systematic_fit:
            invvar = 1./(model.oidata.visamp.error*model.oidata.visamp.error + model.oidata.systematic_prior*model.oidata.systematic_prior)
        else:
            invvar = model.oidata.visamp._invvar
        if chi2:
            quality += ((model.oidata.visamp.data - visamp)**2*invvar).sum()
        else:
            quality += ((model.oidata.visamp.data - visamp)**2*invvar - np.log(invvar)).sum()
    if model.oidata.t3phi:
        if model.oidata.systematic_fit:
            invvar = 1./(model.oidata.t3phi.error*model.oidata.t3phi.error + model.oidata.systematic_prior*model.oidata.systematic_prior)
        else:
            invvar = model.oidata.t3phi._invvar
        if chi2:
            quality += ((model.oidata.t3phi.data - t3phi)**2*invvar).sum()
        else:
            quality += ((model.oidata.t3phi.data - t3phi)**2*invvar - np.log(invvar)).sum()
    if model.oidata.t3amp:
        if model.oidata.systematic_fit:
            invvar = 1./(model.oidata.t3amp.error*model.oidata.t3amp.error + model.oidata.systematic_prior*model.oidata.systematic_prior)
        else:
            invvar = model.oidata.t3amp._invvar
        if chi2:
            quality += ((model.oidata.t3amp.data - t3amp)**2*invvar).sum()
        else:
            quality += ((model.oidata.t3amp.data - t3amp)**2*invvar - np.log(invvar)).sum()

    ### !!! hard-coded shit to fit black-body stuff
    #khi_temp = (core.ratio_bb_flux(1.65*1e-6, params[0], params[4], model._objs[0].diam, model._objs[1].diam) - 1.65)**2*11.111 - 2.40794560
    #khi_temp = (core.ratio_bb_flux(1.65*1e-6, np.random.normal(model._objs[0].temp, 2000), params[3], model._objs[0].diam, model._objs[1].diam) - 1.65)**2*11.111 - 2.40794560
    #quality += khi_temp*model.oidata.vis2.data.size

    if chi2:
        return quality
    else:
        return ln_prior - 0.5*quality # log like


def tweakparams(model, params):
    #Hmag = 5.44-0.62470279
    Rmag = 4.5274 # obtained from V and R mag estimates, extinction corrected
    model._objs[0].diam = soif.core.mag2diam(Rmag - 2.5*np.log10(1./(1. + 1./params[2] + 1./params[4])), 'R', model._objs[0].temp)
    model._objs[1].diam = soif.core.mag2diam(Rmag - 2.5*np.log10((1./params[2])/(1. + 1./params[2] + 1./params[4])), 'R', params[3])
    #model._objs[0].diam = soif.core.mag2diam(Rmag - 2.5*np.log10(1./(1. + 1./params[3] + 1./params[5])), 'R', params[0])
    #model._objs[1].diam = soif.core.mag2diam(Rmag - 2.5*np.log10((1./params[3])/(1. + 1./params[3] + 1./params[5])), 'R', params[4])




#!/usr/bin/env python
# -*- coding: utf-8 -*-

###############################################################################
#  
#  SOIF - Sofware for Optical Interferometry fitting
#  Copyright (C) 2016  Guillaume Schworer
#  
#  This program is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#  
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#  
#  You should have received a copy of the GNU General Public License
#  along with this program.  If not, see <http://www.gnu.org/licenses/>.
#  
#  For any information, bug report, idea, donation, hug, beer, please contact
#    guillaume.schworer@obspm.fr
#
###############################################################################

raise ImportError

import numpy as np
from .oidata import Oidata
from .oimodel import Oimodel
from .oifiting import Oifiting
from .oiunitmodels import *
try:
    import astropy.io.fits as pf
except ImportError:
    import pyfits as pf

from MCres import FakeSampler

__all__ = ['load', 'loaddata', 'loadmodel', 'loadobj']

# known bug, doesn't keep the dataflag mask correct at loading
# doesn't save the prior type, just the lower and higher bounds
ext = '.oif.fits'
card = 'EXT'


def loaddata(filename):
    if filename.find(ext)==-1: filename += ext
    data = None
    hdulist = pf.open(filename)
    # extract data hdu per hdu
    for item in hdulist:
        if item.header.get(card)=='DATA':
            if data is None: # first set
                data = Oidata(src=item.header.get('SRC'), flatten=item.header.get('FLATTEN'), degrees=item.header.get('DEGREES'), significant_figures=item.header.get('SIG_FIG'), **{item.header.get('DATATYPE'):item.data})
            else:
                data.addData(src=item.header.get('SRC'), flatten=item.header.get('FLATTEN'), degrees=item.header.get('DEGREES'), significant_figures=item.header.get('SIG_FIG'), **{item.header.get('DATATYPE'):item.data})
    # apply data mode per mode
    for item in hdulist:
        if item.header.get(card)=='DATAMASK':
            theoldmask = getattr(data, item.header.get('DATATYPE')).mask
            if theoldmask.shape != item.data.shape:
                print("Error while applying data mask on "+item.header.get('DATATYPE'))
            else:
                getattr(data, item.header.get('DATATYPE')).mask = np.logical_or(theoldmask, item.data.astype(bool))
    return data


def loadmodel(filename):
    if filename.find(ext)==-1: filename += ext
    return Oimodel(loaddata(filename), loadobj(filename))


def load(filename):
    cardres = 'MCRES'
    try:
        hdulist = pf.open(filename)
    except IOError:
        if filename.find(ext)==-1: filename += ext
        hdulist = pf.open(filename)
    for item in hdulist:
        if item.header.get(card)==cardres:
            sampler = FakeSampler(item.data[:,:-1], item.data[:,-1])
    return Oifiting(sampler=sampler, model=loadmodel(filename))


def loadobj(filename, name=None):
    if filename.find(ext)==-1: filename += ext
    data = []
    hdulist = pf.open(filename)

    for item in hdulist:
        if item.header.get(card) == 'OBJ':
            if name is None or str(name).upper()==item.header.get('NAME').upper():
                priors = {}
                kwargs = {}
                for i in range(item.header.get('NVALUE')):
                    key = item.header.get('VALUE'+str(i))
                    if item.header.get(key)!='NONE':
                        priors[key] = item.header.get(key)
                bounds = {}
                for i in range(item.header.get('NPARAM')):
                    key = item.header.get('PARAM'+str(i))
                    L = item.header.get(key+'L')
                    H = item.header.get(key+'H')
                    if item.header.get(key)!='NONE':
                        priors[key] = item.header.get(key)
                    bounds[key] = [L, H]
                for i in range(item.header.get('NSAVE')):
                    key = item.header.get('SAVE'+str(i))
                    vv = item.header.get(key)
                    if vv != 'ARRAY':
                        kwargs[key] = vv
                    else:
                        uniq = item.header.get('KEY'+str(i))
                        for itemuniq in hdulist:
                            if itemuniq.header.get(card) == uniq:
                                kwargs[key] = itemuniq.data

                data.append(globals()[item.header.get('TYP')](name=item.header.get('NAME'), priors=priors, bounds=bounds, **kwargs))
    return data



#!/usr/bin/env python
# -*- coding: utf-8 -*-

###############################################################################
#  
#  SOIF - Sofware for Optical Interferometry fitting
#  Copyright (C) 2016  Guillaume Schworer
#  
#  This program is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#  
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#  
#  You should have received a copy of the GNU General Public License
#  along with this program.  If not, see <http://www.gnu.org/licenses/>.
#  
#  For any information, bug report, idea, donation, hug, beer, please contact
#    guillaume.schworer@obspm.fr
#
###############################################################################

import matplotlib.pyplot as plt
try:
    from emcee import EnsembleSampler as emceeEnsembleSampler
    EMCEE = True
except ImportError:
    EMCEE = False
from time import sleep, time

from .oimodel import standardLikelihood

from . import oiexception as exc
from . import core
np = core.np

from MCres import MCres
from patiencebar import Patiencebar

__all__ = ['Oifiting']


class Oifiting(MCres):
    def __init__(self, model, nwalkers=100, niters=500, burnInIts=100, threads=1, customlike=None, **kwargs):
        self.raiseError = bool(kwargs.pop('raiseError', True))
        if not model._hasdata:
            if exc.raiseIt(exc.NoDataModel, self.raiseError):
                return
        if 'sampler' in kwargs.keys():
            self._init(sampler=kwargs.pop('sampler'), paramstr=self.model.paramstr, nwalkers=self.nwalkers,
                       niters=self.niters, burnInIts=self.burnInIts)
        elif EMCEE:
            self.model = model
            self.nwalkers = int(nwalkers)
            self.niters = 0
            self.burnInIts = int(0 if burnInIts is None else burnInIts)
            self.sampler = emceeEnsembleSampler(self.nwalkers, self.model.nparams, standardLikelihood,
                                                args=[self.model, customlike, kwargs], threads=max(1, int(threads)))
            self.run(niters=int(niters), burnInIts=self.burnInIts)

    def _init(self, sampler, paramstr=None, nwalkers=None, niters=None, burnInIts=None):
        super(Oifiting, self)._init(sampler=self.sampler, paramstr=self.model.paramstr, nwalkers=self.nwalkers,
                                    niters=self.niters, burnInIts=self.burnInIts)

    def _info(self):
        return core.font.blue+"<SOIF Fit>{}\n {:d} walkers, {:d} burn-in, {:d} iters\n{}".format(core.font.normal, self.nwalkers, self.burnInIts, self.niters, str(self.model))

    def run(self, niters, burnInIts=0):
        '''
        Start a MCMC simulation on as many CPUs as threads parameter. Use customlike parameter to use a custom
        likelihood function. This function must accept modeledData (ndarray) and model (Oifiting) as input parameters.
        It will also be given the kwargs that you give to this runmc function.
        
            e.g.: mycustomlike(modeledData, model, **kwargs)
        '''
        if not EMCEE:
            if exc.raiseIt(exc.NoEMCEE, self.raiseError):
                return
        if self.model.nparams == 0:
            if exc.raiseIt(exc.AllParamsSet, self.raiseError):
                return
        self.niters = getattr(self, "niters", 0) + int(niters)

        pb = Patiencebar()
        print("{}Running emcee{}".format(core.font.green, core.font.normal))
        t0 = time()

        if burnInIts > 0:
            pb.reset(valmax=int(burnInIts), title="Burn in")
            p0 = [self.model.getP0() for i in range(self.nwalkers)]
            crackit = self.sampler.sample(p0, iterations=burnInIts, storechain=False)
            for res in crackit:
                pb.update()
        else:  # no burn in
            res = [getattr(self, "_p0", None), None, getattr(self, "_rstate0", None)]  # try init where we where
            # if first simu, no burn, was requested at all
            if res[0] is None:
                res[0] = [self.model.getP0() for i in range(self.nwalkers)]

        pb.reset(valmax=int(niters), title="Sampling")
        self.sampler.reset()

        crackit = self.sampler.sample(res[0], iterations=int(niters), rstate0=res[2])
        for res in crackit:
            pb.update()

        sleep(0.1)
        print("Time elapsed = {:.2f}s".format(time()-t0))
        print("Mean acceptance rate: {:.3f}".format(self.sampler.acceptance_fraction.mean()))

        self._p0 = res[0]
        self._rstate0 = res[2]
        self._init(sampler=self.sampler, paramstr=self.model.paramstr, nwalkers=self.nwalkers, niters=self.niters, burnInIts=self.burnInIts)

    def MCmap(self, param_x, param_y, bin_x=50, bin_y=50, cmap="jet", cm_min=None, cm_max=None, axescolor='w', polar=False, showmax=True, radec=False, **kwargs):
        """
        Return a 2D histogram of the MC chain, showing the walker density per bin
        """
        if radec:
            fig, ax = core.astroskyplot(bin_y[-1], polar=polar, unit='mas')
        else:
            if polar:
                fig, ax = plt.subplots(subplot_kw={'projection': 'polar'})
                ax.set_theta_zero_location('N')
            else:
                fig, ax = plt.subplots()
        self._map(param_x=param_x, param_y=param_y, fig=fig, ax=ax, bin_x=bin_x, bin_y=bin_y, cmap=cmap, cm_min=cm_min, cm_max=cm_max, axescolor=axescolor, polar=polar, showmax=showmax, **kwargs)

    def Pbmap(self, param_x, param_y, bin_x=50, bin_y=50, cmap="jet", cm_min=None, cm_max=None, axescolor='w', polar=False, showmax=True, radec=False, **kwargs):
        """
        Return a 2D histogram of the MC chain, showing the best loglikelihood per bin
        """
        if radec:
            fig, ax = core.astroskyplot(bin_y[-1], polar=polar, unit='mas')
        else:
            if polar:
                fig, ax = plt.subplots(subplot_kw={'projection': 'polar'})
                ax.set_theta_zero_location('N')
            else:
                fig, ax = plt.subplots()
        self._map(param_x=param_x, param_y=param_y, fig=fig, ax=ax, data=self.lnprob.compressed(), method=np.max, bin_x=bin_x, bin_y=bin_y, cmap=cmap, cm_min=cm_min, cm_max=cm_max, axescolor=axescolor, polar=polar, showmax=showmax, **kwargs)

    def statut(self, params=None, customlike=None, **kwargs):
        if params is None:
            params = self.best
        self.model.statut(params=params, customlike=customlike)

    def uvimage(self, blmax=None, wl=None, params=None, datatype='Data', nbpts=101, withdata=True, cmap='jet', cm_min=None, cm_max=None, ret=False):
        """
        Datatype in ['vis2', 'phase', 'vis']
        """
        allmodes = ['vis2', 'phase', 'vis']
        calctrick = {'vis2': core.abs2, 'phase': np.angle, 'vis': np.abs}
        cm_min_orig = cm_min
        cm_max_orig = cm_max
        if datatype.lower() == 'all':
            datatype = allmodes
        if datatype.lower() == 'data':
            datatype = []
            if self.model.oidata.vis2: datatype.append('vis2')
            if self.model.oidata.visamp or self.model.oidata.t3amp: datatype.append('vis')
            if self.model.oidata.t3phi or self.model.oidata.visphi: datatype.append('phase')
        elif np.iterable(datatype) == 1 and not isinstance(datatype, str):
            datatype = [str(dum).lower() for dum in datatype if dum.lower() in allmodes]
        else:
            datatype = [str(dum).lower() for dum in [str(datatype)] if dum.lower() in allmodes]
        if blmax is None:
            blmax = np.hypot(self.model.oidata.uvwl['v'], self.model.oidata.uvwl['u']).max()
        if wl is None:
            wl = self.model.oidata.uvwl['wl'].min()
        if params is None:
            params = self.best
        datashown = self.model.compuvimage(blmax=blmax, wl=wl, params=params, nbpts=nbpts)
        if ret:
            retval = []
        for index, datatypeloop in enumerate(datatype):
            data = calctrick[datatypeloop](datashown)
            if ret:
                retval.append(data)
            if cm_min_orig is None:
                cm_min = data.min()
            if cm_max_orig is None:
                cm_max = data.max()
            cmap, norm, mappable = core.colorbar(cmap=cmap, cm_min=cm_min, cm_max=cm_max)
            thefig, ax = core.astroskyplot(blmax, polar=False, unit='m')
            ax.matshow(data, origin='lower', extent=[-blmax, blmax, -blmax, blmax], cmap=cmap, norm=norm)
            if withdata:
                thefig.axes.scatter(getattr(self.model.oidata, datatypeloop).u, getattr(self.model.oidata, datatypeloop).v, c=getattr(self.model.oidata, datatypeloop).data, cmap=cmap, norm=norm)
            plt.colorbar(mappable)
            thefig.axes.set_title(datatypeloop)
        if ret:
            return retval

    def image(self, params=None, sepmax=None, wl=None, masperpx=None, nbpts=101, cmap='jet', cm_min=None, cm_max=None, ret=False, visu=None, **visuargs):
        if params is None:
            params = self.best
        return self.model.image(params=params, sepmax=sepmax, wl=wl, masperpx=masperpx, nbpts=nbpts, cmap=cmap, cm_min=cm_min, cm_max=cm_max, ret=ret, visu=visu, **visuargs)

    def imagefft(self, sepmax=None, wl=None, params=None, nbpts=101, cmap='jet', cm_min=None, cm_max=None, ret=False, visu=None, **visuargs):
        if sepmax is None:
            sepmax = 1000*nbpts/(2*np.hypot(self.model.oidata.uvwl['v'], self.model.oidata.uvwl['u']).max())
        if wl is None:
            wl = self.model.oidata.uvwl['wl'].min()
        if params is None:
            params = self.best
        if visu is None:
            visu = np.abs
        #toplot = visu(self.model.compimage(sepmax=sepmax, wl=wl, params=params, nbpts=nbpts), **visuargs)
        toplot = self.model.compuvimage(blmax=nbpts*2*sepmax*core.MAS2RAD/np.pi/wl, wl=wl, params=params, nbpts=nbpts)
        toplot = np.abs(np.fft.fftshift(np.fft.ifft2(toplot)))
        if cm_min is None:
            cm_min = toplot.min()
        if cm_max is None:
            cm_max = toplot.max()
        cmap, norm, mappable = core.colorbar(cmap=cmap, cm_min=cm_min, cm_max=cm_max)
        thefig, ax = core.astroskyplot(sepmax, polar=False, unit='mas')
        ax.matshow(toplot, origin='lower', extent=[-sepmax, sepmax, -sepmax, sepmax], cmap=cmap, norm=norm)
        plt.colorbar(mappable)
        if ret:
            return toplot

    def residual(self, params=None, c=None, cmap='jet', cm_min=None, cm_max=None, datatype='All'):
        if params is None:
            params = self.best
        self.model.residual(params=params, c=c, cmap=cmap, cm_min=cm_min, cm_max=cm_max, datatype=datatype)

    def save(self, name, clobber=False):
        name = self.model.save(name, clobber=clobber)
        super(Oifiting, self).save(name=name, clobber=clobber, append=True)
        return name


#! /usr/bin/env python
# -*- coding: utf-8 -*-

from sys import argv, exit
import os, re

if "upl" in argv[1:]:
    os.system("python setup.py register -r pypi")
    os.system("python setup.py sdist upload -r pypi")
    exit()

m = open(os.path.join(os.path.dirname(os.path.abspath(__file__)), "soif", "_version.py")).read()
version = re.findall(r"__version__ *= *\"(.*?)\"", m)[0]

try:
    from setuptools import setup
    setup
except ImportError:
    from distutils.core import setup
    setup

setup(
    name = "soif",
    version = version,
    author = "Guillaume Schworer",
    author_email = "guillaume.schworer@obspm.fr",
    packages = ["soif"],
    url = "https://github.com/ceyzeriat/soif/",
    license = "GNU General Public License v3 or later (GPLv3+)",
    description = "Software for Optical Interferometry Fitting",
    long_description = open("README.rst").read() + "\n\n"
                    + "Changelog\n"
                    + "---------\n\n"
                    + open("HISTORY.rst").read(),
    include_package_data = True,
    install_requires = ["numpy>1.6.1","emcee","corner","patiencebar","MCres"],
    download_url = 'https://github.com/ceyzeriat/soif/tree/master/dist',
    keywords = ['astronomy','interferometry','data','processing','reduction','model','fitting','optical'],
    package_data={"": ["README.rst", "LICENSE", "HISTORY.rst"]},
    classifiers = [
        "Development Status :: 4 - Beta",
        "Environment :: Console",
        "Intended Audience :: Developers",
        'Intended Audience :: Education',
        "Intended Audience :: Science/Research",
        "License :: OSI Approved :: GNU General Public License v3 or later (GPLv3+)",
        "Operating System :: OS Independent",
        "Programming Language :: Python",
        "Topic :: Documentation :: Sphinx",
        "Topic :: Scientific/Engineering :: Physics",
        "Topic :: Scientific/Engineering :: Astronomy"
    ],
)




#!/usr/bin/env python
# -*- coding: utf-8 -*-

###############################################################################
#
#  SOIF - Sofware for Optical Interferometry fitting
#  Copyright (C) 2016  Guillaume Schworer
#
#  This program is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
#  For any information, bug report, idea, donation, hug, beer, please contact
#    guillaume.schworer@obspm.fr
#
###############################################################################

from time import strftime
try:
    import astropy.io.fits as pf
except ImportError:
    import pyfits as pf

from . import oiexception as exc
from .oidataempty import OidataEmpty
from .oidata import Oidata
from . import core
np = core.np

__all__ = ['Oifits']


class Oifits(object):
    """
    This class opens, reads and sorts data contained in the file
    'src' (oifits format).
    """
    def __init__(self, src, datafilter, wl=(None, None), erb_sigma=None,
                 sigma_erb=None, systematic_prior=None, systematic_bounds=(),
                 flatten=False, degree=True, significant_figures=5, **kwargs):
        self.raiseError = bool(kwargs.pop('raiseError', True))
        # initialize empty data
        for key in core.DATAKEYSLOWER:
            setattr(self, key, OidataEmpty(key))

        self.erb_sigma = core.ident if erb_sigma is None else erb_sigma
        if not callable(self.erb_sigma):
            if exc.raiseIt(exc.NotCallable, self.raiseError, fct="erb_sigma"):
                return
        self.sigma_erb = core.ident if sigma_erb is None else sigma_erb
        if not callable(self.sigma_erb):
            if exc.raiseIt(exc.NotCallable, self.raiseError, fct="sigma_erb"):
                return
        self.systematic_bounds = None \
            if len(systematic_bounds) <= 2  \
            else list(map(float, list(systematic_bounds)[:2]))
        self.systematic_prior = None \
            if systematic_prior is None \
            else float(systematic_prior)
        self._systematic_prior = self.systematic_prior
        # init parameters
        self.addData(src=str(src), datafilter=dict(datafilter),
                     flatten=bool(flatten), degree=bool(degree),
                     significant_figures=int(significant_figures),
                     wl=list(wl)[:2], noupdate=True, **kwargs)
        self.update()

    def _info(self):
        txt = "\n".join([" {}".format(getattr(self, key))
                         for key in core.DATAKEYSLOWER
                         if getattr(self, key).useit])
        txt = " No data" if txt == "" else txt
        txt = "{}<SOIF Data>{}\n{}".format(core.font.blue,
                                           core.font.normal,
                                           txt)
        return txt

    def __repr__(self):
        return self._info()

    __str__ = __repr__

    def addData(self, src, datafilter=None, flatten=False, degree=True,
                significant_figures=5, wl=(None, None), **kwargs):
        if datafilter is None:
            datafilter = {}
        if isinstance(src, Oifits):
            thedata = src
            for datatype in core.DATAKEYSLOWER:
                if not getattr(thedata, datatype):
                    continue
                if getattr(self, datatype):
                    getattr(self, datatype)._addData(
                                                getattr(thedata, datatype),
                                                flatten=flatten,
                                                **kwargs)
                else:
                    setattr(self, datatype, getattr(thedata, datatype))
        else:
            whichdata = datafilter.get('data',
                                       {'data': {'VIS2': True,
                                                 'T3PHI': True,
                                                 'T3AMP': True,
                                                 'VISPHI': True,
                                                 'VISAMP': True}})
            hdus = pf.open(src)  # open
            hduwlidx = core.hduWlindex(hdus)
            if not hduwlidx:
                if exc.raiseIt(exc.NoWavelengthTable,
                               self.raiseError,
                               src=src):
                    return
            # get wl sorted
            wl = [float(wl[0] if wl[0] is not None else -np.inf),
                  float(wl[1] if wl[1] is not None else np.inf)]
            allwl = hdus[hduwlidx].data[core.KEYSWL['wl']]
            wlindices = np.arange(allwl.size)[((allwl >= wl[0])
                                               & (allwl < wl[1]))]
            # for each datafilter
            for idx, indices in datafilter.items():
                if not isinstance(idx, int) or np.size(indices) == 0:
                    continue
                # if real data
                if core.hduToDataType(hdus[idx]) is not None:
                    hduextname = hdus[idx].header['EXTNAME']
                    for datatype in core.ALLDATAEXTNAMES[hduextname]:
                        if not whichdata.get(datatype.upper(), True):
                            continue
                        thedata = Oidata(
                            src=src, hduidx=idx, datatype=datatype,
                            hduwlidx=hduwlidx, indices=indices,
                            wlindices=wlindices, degree=degree,
                            flatten=flatten,
                            significant_figures=significant_figures, **kwargs)
                        if getattr(self, datatype.lower()):
                            getattr(self, datatype.lower())._addData(
                                                            thedata,
                                                            flatten=flatten,
                                                            **kwargs)
                        else:
                            setattr(self, datatype.lower(), thedata)
                else:
                    hdus.close()
                    if exc.raiseIt(exc.NotADataHdu,
                                   self.raiseError,
                                   idx=idx,
                                   src=str(src)):
                        return
            hdus.close()
        if not kwargs.pop('noupdate', False):
            self.update()

    def flatten(self):
        """
        Flattens all data contained in the Oidata object. This can be
        useful in order to add several bits of data that do not have
        the same shapes
        """
        for key in core.DATAKEYSLOWER:
            if getattr(self, key):
                getattr(self, key).flatten()
        self.update()

    def update(self):
        """
        Updates all data contained in the Oidata object
        """
        for key in core.DATAKEYSLOWER:
            if getattr(self, key):
                getattr(self, key).update()

        funkydtypeint = [('u', int), ('v', int), ('wl', int)]
        funkydtype = [('u', np.float32), ('v', np.float32), ('wl', np.float32)]

        # get uvwl sets as integer to extract uniques
        unique_uvwl = np.zeros(0, dtype=funkydtypeint)
        for key in core.DATAKEYSLOWER:
            thedata = getattr(self, key)
            if thedata:
                dum = np.zeros(thedata.shapeuv, dtype=funkydtypeint)
                dum['u'] = core.round_fig(x=thedata.u,
                                          n=thedata.significant_figures,
                                          retint=True)
                dum['v'] = core.round_fig(x=thedata.v,
                                          n=thedata.significant_figures,
                                          retint=True)
                dum['wl'] = core.round_fig(x=thedata.wl,
                                           n=thedata.significant_figures,
                                           retint=True)
                # deal with symmetry
                inv = (dum['u'] < 0)
                dum['u'][inv] *= -1
                dum['v'][inv] *= -1
                # stack
                unique_uvwl = np.hstack((unique_uvwl, dum.flatten()))
        if core.OLDNUMPY:
            uvwlind = np.unique((unique_uvwl['u']+1j*unique_uvwl['v'])
                                / unique_uvwl['wl'],
                                return_index=True)[1]
        else:
            uvwlind = np.unique(unique_uvwl, return_index=True)[1]
        # get uvwl sets as floats for calculations
        unique_uvwl = np.zeros(0, dtype=funkydtype)
        for key in core.DATAKEYSLOWER:
            thedata = getattr(self, key)
            if thedata:
                dum = np.zeros(thedata.shapeuv, dtype=funkydtype)
                dum['u'] = core.round_fig(x=thedata.u,
                                          n=thedata.significant_figures)
                dum['v'] = core.round_fig(x=thedata.v,
                                          n=thedata.significant_figures)
                dum['wl'] = core.round_fig(x=thedata.wl,
                                           n=thedata.significant_figures)
                # simplify symmetry
                inv = (dum['u'] < 0)
                dum['u'][inv] *= -1
                dum['v'][inv] *= -1
                # save the phase symmetry
                if thedata.is_angle:
                    thedata._phasesign = 1-inv*2
                # stack
                unique_uvwl = np.hstack((unique_uvwl, dum.flatten()))
                thedata._ind = dum.copy()
        # extract uniques
        dum = np.zeros(uvwlind.shape, dtype=funkydtype)
        dum['u'] = unique_uvwl['u'][uvwlind]
        dum['v'] = unique_uvwl['v'][uvwlind]
        dum['wl'] = unique_uvwl['wl'][uvwlind]
        unique_uvwl = dum
        # save indices
        for key in core.DATAKEYSLOWER:
            thedata = getattr(self, key)
            if thedata:
                dum = np.zeros(thedata._ind.shape, dtype='int')
                for i, v in enumerate(unique_uvwl):
                    dum[thedata._ind == v] = i
                thedata._ind = dum.copy()
        # prepare pre-processed uniques
        self.uvwl = {'u': unique_uvwl['u'],
                     'v': unique_uvwl['v'],
                     'wl': unique_uvwl['wl'],
                     'blwl': np.hypot(unique_uvwl['u'], unique_uvwl['v'])
                             / unique_uvwl['wl']}
        self._wlmin = self.uvwl['wl'].min()
        self._wlmax = self.uvwl['wl'].max()
        self._wlspan = self._wlmax - self._wlmin

    @property
    def systematic_fit(self):
        return isinstance(self.systematic_bounds, list)

    @systematic_fit.setter
    def systematic_fit(self, value):
        exc.raiseIt(exc.ReadOnly, self.raiseError, attr="systematic_fit")

    def systematic_p0(self):
        if self.systematic_fit:
            randomizer = core.gen_generator()
            return randomizer.uniform(low=self.systematic_bounds[0],
                                      high=self.systematic_bounds[1])
        else:
            exc.raiseIt(exc.NoSystematicsFit, self.raiseError)

    def remorph(self, viscomp):
        ret = []
        for key in core.DATAKEYSLOWER:
            thedata = getattr(self, key)
            if thedata.useit:
                retdbl = isinstance(viscomp, (tuple, list))
                if retdbl:
                    viscomp, flx = viscomp
                if thedata.is_t3:
                    fct = core.FCTVISCOMP[key.upper()]
                    if thedata.is_angle:  # t3phi
                        dum = (fct(viscomp)[thedata._ind]
                               * thedata._phasesign).sum(-1)
                    else:  # t3amp
                        dum = (fct(viscomp[thedata._ind])).prod(-1)
                else:
                    dum = core.FCTVISCOMP[key.upper()](viscomp)[thedata._ind]
                    if thedata.is_angle:  # visphi
                        dum *= thedata._phasesign
                if retdbl:
                    ret.append((dum, flx[thedata._ind]))
                else:
                    ret.append(dum)
            else:
                ret.append(None)
        return ret

    def save(self, filename, append=False, clobber=False):
        ext = '.oif.fits'
        if filename.find(ext) == -1:
            filename += ext
        if append:
            hdulist = pf.open(filename, mode='append')
        else:
            hdulist = pf.HDUList()
        allmodes = ['vis2', 't3phi', 't3amp', 'visphi', 'visamp']
        for i in range(len(self._input_src)):
            for mode in allmodes:
                if getattr(self, "_has"+mode):
                    hdu = pf.PrimaryHDU()
                    hdu.header.set('EXT',
                                   'DATA',
                                   comment='Type of information in the HDU')
                    hdu.header.set('DATE',
                                   strftime('%Y%m%dT%H%M%S'),
                                   comment='Creation Date')
                    hdu.header.set('DATAFILE', i, comment='Data file number')
                    hdu.header.set('SRC',
                                   str(self._input_src[i]),
                                   comment='Path to the datafile')
                    hdu.header.set('DATATYPE',
                                   mode,
                                   comment='Type of data from'
                                           ' datafile {:d}'.format(i))
                    hdu.header.set('FLATTEN',
                                   bool(self._input_flatten[i]),
                                   comment='Should the data be flatten')
                    hdu.header.set('SIG_FIG',
                                   int(self.significant_figures),
                                   comment='Significant figures for'
                                           'u,v coordinates')
                    if mode in ['visphi', 't3phi']:
                        hdu.header.set('DEGREES',
                                       bool(self._input_degrees[i]),
                                       comment='Is datafile in degrees')
                    hdu.data = np.ravel(getattr(self, "_input_"+mode)[i])

                    hdu.header.add_comment('Measurement indices to be imported'
                                           ' from datafile for the given'
                                           ' datatype.')
                    hdu.header.add_comment('Written by Guillaume SCHWORER')
                    hdulist.append(hdu)
        for mode in allmodes:
            if getattr(self, "_has"+mode):
                hdu = pf.PrimaryHDU()
                hdu.header.set('EXT',
                               'DATAMASK',
                               comment='Type of information in the HDU')
                hdu.header.set('DATE',
                               strftime('%Y%m%dT%H%M%S'),
                               comment='Creation Date')
                hdu.header.set('DATATYPE',
                               mode,
                               comment='Type of data from datafile '+str(i))
                hdu.data = getattr(self, mode).mask.astype(np.uint8)
                hdu.header.add_comment('Data mask for the given datatype for'
                                       ' all data contained in the different'
                                       ' files.')
                hdu.header.add_comment('Written by Guillaume SCHWORER')
                hdulist.append(hdu)

        if append:
            hdulist.flush()
            hdulist.close()
        else:
            hdulist.writeto(filename, clobber=clobber)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

###############################################################################
#
#  SOIF - Sofware for Optical Interferometry fitting
#  Copyright (C) 2016  Guillaume Schworer
#
#  This program is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
#  For any information, bug report, idea, donation, hug, beer, please contact
#    guillaume.schworer@obspm.fr
#
###############################################################################


# import matplotlib.pyplot as _plt
from . import oiexception as exc
from . import core

__all__ = []


class OidataEmpty(object):
    def __init__(self, datatype, **kwargs):
        self.raiseError = bool(kwargs.pop('raiseError', True))
        self.datatype = str(datatype).upper()
        if self.datatype not in core.DATAKEYSUPPER:
            if exc.raiseIt(exc.InvalidDataType,
                           self.raiseError,
                           datatype=self.datatype):
                return
        self._has = False
        self._useit = False

    def _info(self):
        return "{} data, None".format(self.datatype)

    def __repr__(self):
        return self._info()

    __str__ = __repr__

    @property
    def useit(self):
        """
        Turn this True or False whether you wan't the fitting to
        include this datatype
        """
        return self._useit and self._has

    @useit.setter
    def useit(self, value):
        self._useit = bool(value)

    def __bool__(self):
        return self._has

    __nonzero__ = __bool__

    def __getitem__(self, key):
        return getattr(self, "_"+key)
